#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <dirent.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "log.h"
#include "camera.h"
#include "devsec.h"
#include "dtk_config.h"

#define USE_NEW_DIR

#ifdef USE_NEW_DIR
char *months[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
			"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
#endif

Camera::Camera(Widget *parent, int x, int y, int w, int h, int cameraNo)
	: Widget(parent, x, y, w, h)
{
	char conffile[1024];
	char sockfile[1024];
	char pidfile[1024];
	char section[128];

	SetColor(0x0, 0x0, 0x0);
	mGlobalROM = 0;
	mGlobalNOM = 0;
	mCameraNo = cameraNo;
	snprintf(conffile, 1024, CAMERA_CONFIG, cameraNo);
	snprintf(sockfile, 1024, CAMERA_SOCKET, cameraNo);
	snprintf(pidfile, 1024, CAMERA_PID, cameraNo);
	snprintf(section, 128, "video%d", cameraNo);

	mConfFile = strdup(conffile);
	mSockFile = strdup(sockfile);
	mPIDFile = strdup(pidfile);
	mSection = strdup(section);

	mLocked = 0;
	mCameraName = NULL;
	mRecording = 0;
	mBrightness = 32767;
	mContrast = 32767;
	mColor = 32767;
	mAlive = 0;
	mActive = 0;

	mStoppedPending = 0;
	mStartWaiting = 0;

	LoadConfig();

	mRemoteW = 1;
	mRemoteH = 1;
	mRemoteDisplay = NULL;
	mRemoteParent = 0;
	mRemoteVisible = 0;
	mRemoteInViewList = 0;

	mRecordLengthTimer = new Timer(5, this, NULL, TimerCallback);
	mRecordLengthTimer->Start();
	AddChild(mRecordLengthTimer);
	mPopup = NULL;
	mBackendPID = 0;
	mSocket = new Socket();
	AddChild(mSocket);
	StartBackend();
}

Camera::~Camera()
{
	if (mConfFile) free((void*)mConfFile);
	if (mSockFile) free((void*)mSockFile);
	if (mPIDFile) free((void*)mPIDFile);
	if (mSection) free((void*)mSection);
	StopBackend();
}

int Camera::SetPopup(Popup *popup)
{
	mPopup = popup;
	return 0;
}

int Camera::EventHandler(TEvent *ev)
{
	int rv = 0;

	rv = Widget::EventHandler(ev);

	if (ev == NULL) return 0;
	if (ev->type == NOEVENT) return 0;

	if (ev->type == TIMEEVENT) {
		if (mSocket->Pending()) {
			SockEvent sev;
			GetNextEvent(&sev);
			switch (sev.type) {
				case MOUSE_BUTTON_EVENT_TYPE:
					if (sev.mousebutton.state == UP)
						return 0;
					if (sev.mousebutton.button != 3)
						return 0;
					if (mPopup) mPopup->Show();
					break;
				case MOTION_DETECT_EVENT_TYPE:
					DoMotionDetect(sev.motiondetect.detected);
					break;
			}
		}
	}
	return rv;
}

int Camera::Resize(int w, int h)
{
	PreviewSize(w, h);
	return Widget::Resize(w, h);
}

int Camera::MoveResize(int x, int y, int w, int h)
{
	PreviewSize(w, h);
	return Widget::MoveResize(x, y, w, h);
}

int Camera::Lock(int lock)
{
	mLocked = lock;
	if (mPopup == NULL) return 0;

	if (mLocked) mPopup->SetEntryText("Unlock View", 0);
	else mPopup->SetEntryText("Lock View  ", 0);
	return 0;
}

int Camera::GlobalROM(int flag)
{
	mGlobalROM = flag;
	if (flag && mAlive) mSocket->Write("MOTION resetmotion\n");
	return 0;
}

int Camera::GlobalNOM(int flag)
{
	mGlobalNOM = flag;
	return 0;
}

int Camera::LoadConfig()
{
	char *cameraName;

	Config::ReadConfig(mSection, mConfFile);

	cameraName = Config::GetCharValue(mSection, "CameraName", NULL);

	if (cameraName == NULL) {
		char buf[64];
		snprintf(buf, 64, "Video%d", mCameraNo);
		cameraName = strdup(buf);
	}

	SetCameraName(cameraName);

	mBoundsX1 = Config::GetIntValue(mSection, "BoundsLeft", 0);
	mBoundsY1 = Config::GetIntValue(mSection, "BoundsTop", 13);
	mBoundsX2 = Config::GetIntValue(mSection, "BoundsRight", 383);
	mBoundsY2 = Config::GetIntValue(mSection, "BoundsBottom", 287);
	mSensitivity = Config::GetIntValue(mSection, "Sensitivity", 160);
	mShowBounds = Config::GetIntValue(mSection, "ShowBounds", 0);
	mShowMotion = Config::GetIntValue(mSection, "ShowMotion", 0);
	mRecordMotion = Config::GetIntValue(mSection, "RecordOnMotion", 0);
	mNotifyMotion = Config::GetIntValue(mSection, "NotifyOnMotion", 0);
	mRecordAlways = Config::GetIntValue(mSection, "RecordAlways", 1);
	mBrightness = Config::GetIntValue(mSection, "Brightness", 32767);
	mContrast = Config::GetIntValue(mSection, "Contrast", 32767);
	mColor = Config::GetIntValue(mSection, "Color", 32767);
	mBitrate = Config::GetIntValue(mSection, "videoBitRate", 500);

	if (mAlive) DumpConfiguration();
	return 0;
}

int Camera::SaveConfig()
{
	Config::SetCharValue(mSection, "CameraName", mCameraName);
	Config::SetIntValue(mSection, "BoundsLeft", mBoundsX1);
	Config::SetIntValue(mSection, "BoundsTop", mBoundsY1);
	Config::SetIntValue(mSection, "BoundsRight", mBoundsX2);
	Config::SetIntValue(mSection, "BoundsBottom", mBoundsY2);
	Config::SetIntValue(mSection, "Sensitivity", mSensitivity);
	Config::SetIntValue(mSection, "ShowBounds", mShowBounds);
	Config::SetIntValue(mSection, "ShowMotion", mShowMotion);
	Config::SetIntValue(mSection, "RecordOnMotion", mRecordMotion);
	Config::SetIntValue(mSection, "NotifyOnMotion", mNotifyMotion);
	Config::SetIntValue(mSection, "RecordAlways", mRecordAlways);
	Config::SetIntValue(mSection, "Brightness", mBrightness);
	Config::SetIntValue(mSection, "Contrast", mContrast);
	Config::SetIntValue(mSection, "Color", mColor);
	Config::SetIntValue(mSection, "videoBitRate", mBitrate);

	Config::WriteConfig(mSection, mConfFile);

	if (mAlive) DumpConfiguration();

	return 0;
}

int Camera::CutStream()
{
	if (mRecording) {
		MpegStop();
		MpegStart();
	}
	return 0;
}

int Camera::GetNextEvent(SockEvent *sev)
{
	char buf[256];
	int val = 0;

	if (sev == NULL) return 0;
	memset(sev, 0, sizeof(SockEvent));

	if (mSocket->Pending() == 0) return 0;

	if (mSocket->Read(buf, 255) < 0) {
		mSocket->Disconnect();
		// We won't connect again. If the backend has died, we will
		// reconnect when we restart it.
		// mSocket->Connect(mSockFile);
		write_log("lost connection with the backend\n");

		return 0;
	}
	if (!strncmp(buf, "EV_KEY_UP", 9)) {
		sscanf(buf, "EV_KEY_UP %d", &val);
		sev->type = KEY_EVENT_TYPE;
		sev->key.state = UP;
		sev->key.ascii = (unsigned char)val;
		return 1;
	} else if (!strncmp(buf, "EV_KEY_DOWN", 11)) {
		sscanf(buf, "EV_KEY_DOWN %d", &val);
		if (val > 1000) return 0;
		sev->type = KEY_EVENT_TYPE;
		sev->key.state = DOWN;
		sev->key.ascii = (unsigned char)val;
		return 1;
	} else if (!strncmp(buf, "EV_MOUSE_UP", 11)) {
		sscanf(buf, "EV_MOUSE_UP %d", &val);
		if (val < 1000) return 0;
		sev->type = MOUSE_BUTTON_EVENT_TYPE;
		sev->mousebutton.state = UP;
		sev->mousebutton.button = val - 1040;
		return 1;
	} else if (!strncmp(buf, "EV_MOUSE_DOWN", 13)) {
		sscanf(buf, "EV_MOUSE_DOWN %d", &val);
		sev->type = MOUSE_BUTTON_EVENT_TYPE;
		sev->mousebutton.state = DOWN;
		sev->mousebutton.button = val;
		return 1;
	} else if (!strncmp(buf, "EV_MOUSE_MOTION", 15)) {
		unsigned int x, y;
		sscanf(buf, "EV_MOUSE_MOTION %d %d", &x, &y);
		sev->type = MOUSE_MOTION_EVENT_TYPE;
		sev->mousemotion.x = x;
		sev->mousemotion.y = y;
		return 1;
	} else if (!strncmp(buf, "EV_MOTION_START", 15)) {
		sev->type = MOTION_DETECT_EVENT_TYPE;
		sev->motiondetect.detected = 1;
		sev->motiondetect.dtime = time(NULL);
		return 1;
	} else if (!strncmp(buf, "EV_MOTION_STOP", 14)) {
		sev->type = MOTION_DETECT_EVENT_TYPE;
		sev->motiondetect.detected = 0;
		sev->motiondetect.dtime = time(NULL);
		return 1;
	} else if (!strncmp(buf, "HELO", 4)) {
		mAlive = 1;
		DumpConfiguration();
	} else if (!strncmp(buf, "ACTIVE", 6)) {
		mActive = 1;
	} else if (!strncmp(buf, "STOPPED", 7)) {
		RenameFile(0);
		mStoppedPending = 0;
		my_system("killall -CONT devsec_helper > /dev/null 2>&1");
		if (mStartWaiting) MpegStart();
	}
	return 0;
}

int Camera::DoMotionDetect(int flag)
{
	if ((mGlobalNOM) && (mNotifyMotion)) {
		if (flag) {
			char buf[128];
			char *phoneNo;

			phoneNo = Config::GetCharValue("global", "PhoneNumber",
					NULL);
			if (phoneNo == NULL) return 0;
			snprintf(buf, 128, "devsec_helper notify %d %s&",
					mCameraNo, phoneNo);
			free((void*)phoneNo);
			my_system(buf);
		}
	}

	if (mRecordAlways  && !mRecording) {
		MpegStart();
		return 0;
	}

	if (mRecordAlways) return 0;

	if (mGlobalROM && (mRecordMotion || mRecording)) {
		if (flag) { 
			MpegStart();
			if (Devsec::sAudibleRecord) XBell(Widget::GetXDisplay(),
					100);
		} else MpegStop();
	}
	return 0;
}

int Camera::SetCameraName(char *name)
{
	char buf[256];
	char oldname[256];
	char newname[256];

	if (name != mCameraName) {
		if (mCameraName) {
			DIR *d;
			snprintf(oldname, 256, STREAM_DIR "/%s", mCameraName);
			snprintf(newname, 256, STREAM_DIR "/%s", name);
			d = opendir(newname);
			if (d != NULL) {
				closedir(d);
				return 0;
			}
			if (rename(oldname, newname) != 0) return 0;
			free((void*)mCameraName);
		}

		mCameraName = strdup(name);
		snprintf(newname, 256, STREAM_FILE, mCameraName);
		snprintf(buf, 256, "STREAM file %s\n", newname);
		if (mAlive) {
			mSocket->Write(buf);
		}
	}
	snprintf(buf, 256, "TEXT text %s /%s/\n", mCameraName,
			(mRecording ? "REC" : "STOP"));
	if (mAlive) mSocket->Write(buf);
	return 0;
}

char *Camera::GetCameraName()
{
	return mCameraName;
}

int Camera::RestartBackend()
{
	char buf[256];
	snprintf(buf, 256, "restarting tsrep3 (%d)", mBackendPID);
	write_log(buf);
	StopBackend();
	StartBackend();
	return 0;
}

int Camera::IsZOMBIE()
{
	int rv = 0;
	char buf[256];
	rv = waitpid(mBackendPID, NULL, WNOHANG);

	if (rv == mBackendPID) {
		snprintf(buf,256,"tsrep3 (%d) has become a zombie",
				mBackendPID);
		write_log(buf);
		return 1;
	}

	if (rv == -1) {
		if (kill(mBackendPID, SIGHUP) < 0) {
			snprintf(buf,256,"tsrep3 (%d) has died",mBackendPID);
			write_log(buf);
			return 1;
		}
	}
	
	return 0;
}

int Camera::V4LBrightness(unsigned int brightness)
{
	char buf[64];

	mBrightness = brightness;
	snprintf(buf, 64, "V4L bright=%u\n", brightness);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::V4LContrast(unsigned int contrast)
{
	char buf[64];

	mContrast = contrast;
	snprintf(buf, 64, "V4L contrast=%u\n", contrast);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::V4LColor(unsigned int color)
{
	char buf[64];

	mColor = color;
	snprintf(buf, 64, "V4L color=%u\n", color);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::MotionBounds(int x1, int y1, int x2, int y2)
{
	char buf[64];

	if (x1 > -1) mBoundsX1 = x1;
	if (y1 > -1) mBoundsY1 = y1;
	if (x2 > -1) mBoundsX2 = x2;
	if (y2 > -1) mBoundsY2 = y2;

	snprintf(buf, 64, "MOTION bounds=%dx%dx%dx%d\n",
			mBoundsX1, mBoundsY1,
			mBoundsX2, mBoundsY2);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::SetSensitivity(int sensitivity)
{
	char buf[64];

	mSensitivity = sensitivity;
	snprintf(buf, 64, "MOTION sensitivity=%d\n", sensitivity);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::ShowBounds(int flag)
{
	char buf[64];

	mShowBounds = flag;
	snprintf(buf, 64, "MOTION showbounds=%d\n", flag);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::ShowMotion(int flag)
{
	char buf[64];

	mShowMotion = flag;
	snprintf(buf, 64, "MOTION showmotion=%d\n", flag);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::SetMotionRecord(int flag)
{
	mRecordMotion = flag;
	if (!flag) MpegStop();
	else if (mAlive) mSocket->Write("MOTION resetmotion\n");
	return 0;
}

int Camera::SetNotification(int flag)
{
	mNotifyMotion = flag;
	return 0;
}

int Camera::RecordAlways(int flag)
{
	mRecordAlways = flag;
	if (flag) MpegStart();
	else MpegStop();
	return 0;
}

int Camera::PreviewSize(int w, int h)
{
	char buf[64];

	snprintf(buf, 64, "PREVIEW local size=%dx%d\n", w, h);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::PreviewParent(Window parent)
{
	char buf[64];

	snprintf(buf, 64, "PREVIEW local parent=%u\n", (unsigned int)parent);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::PreviewDisplay(char *display)
{
	char buf[64];

	snprintf(buf, 64, "PREVIEW local display=%s\n", display);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::PreviewShow()
{
	if (mAlive) mSocket->Write("PREVIEW local show\n");
	return 0;
}

int Camera::PreviewHide()
{
	if (mAlive) mSocket->Write("PREVIEW local hide\n");
	return 0;
}

int Camera::RemoteSize(int w, int h)
{
	char buf[64];

	mRemoteW = w;
	mRemoteH = h;
	snprintf(buf, 64, "PREVIEW remote size=%dx%d\n", w, h);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::RemoteParent(Window parent)
{
	char buf[64];

	mRemoteParent = parent;
	snprintf(buf, 64, "PREVIEW remote parent=%u\n", (unsigned int)parent);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::RemoteDisplay(char *display)
{
	char buf[64];

	if ((display != mRemoteDisplay) && (mRemoteDisplay)) {
		free((void*)mRemoteDisplay);
		mRemoteDisplay = (display ? strdup(display) : NULL);
	}
	snprintf(buf, 64, "PREVIEW remote display=%s\n", display);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::RemoteShow()
{
	mRemoteVisible = 1;
	if (mAlive) mSocket->Write("PREVIEW remote show\n");
	return 0;
}

int Camera::RemoteHide()
{
	mRemoteVisible = 0;
	if (mAlive) mSocket->Write("PREVIEW remote hide\n");
	return 0;
}

int Camera::SetRemoteInViewList(int flag)
{
	mRemoteInViewList = flag;
	return 0;
}

int Camera::GetRemoteInViewList()
{
	return mRemoteInViewList;
}

int Camera::MpegStart()
{
	char buf[256];

	snprintf(buf, 256, "%s/%s", STREAM_DIR, mCameraName);
	if (access(buf, R_OK)) {
		snprintf(buf, 256, "mkdir -p '%s/%s'", STREAM_DIR, mCameraName);
		my_system(buf);
	}
	mStartWaiting = 1;
	if (mStoppedPending) {
		return 0;
	}

	mPrevSize = 0;
	mStartTime = time(NULL);
	mStartWaiting = 0;

	mBackendPrio = getpriority(PRIO_PROCESS, mBackendPID);
	setpriority(PRIO_PROCESS, mBackendPID, -20);
	if (mAlive) mSocket->Write("MPEG start\n");
	snprintf(buf, 256, "TEXT text %s /REC/\n", mCameraName);
	if (mAlive) mSocket->Write(buf);
	mRecording = 1;
	return 0;
}

int Camera::MpegStop()
{
	char buf[64];

	if (!mRecording) return 0;

	my_system("killall -STOP devsec_helper > /dev/null 2>&1");
	mStoppedPending = 1;
	snprintf(buf, 64, "TEXT text %s /STOP/\n", mCameraName);
	if (mAlive) mSocket->Write(buf);
	if (mAlive) mSocket->Write("MPEG stop\n");
	mRecording = 0;
	return 0;
}

int Camera::SetBitrate(unsigned int bitrate)
{
	char buf[64];

	mBitrate = bitrate;
	snprintf(buf, 64, "MPEG bitrate=%u\n", bitrate);
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::UnicastDisplay(char *addr, char *port)
{
	char buf[256];
	if ((addr == NULL) || (port == NULL)) {
		snprintf(buf, 256, "UNICAST OFF\n");
	} else {
		snprintf(buf, 256, "UNICAST %s %s\n", addr, port);
	}
	if (mAlive) mSocket->Write(buf);
	return 0;
}

int Camera::OnTimerCallback()
{
	char streamName[256];
	struct stat buf;
	time_t now = time(NULL);

	if (!mRecording) return 0;

	snprintf(streamName, 256, STREAM_FILE, mCameraName);
	if (stat(streamName, &buf) == -1) return 0;

	if ((buf.st_size == 24) && (mPrevSize == 24)) {
		char tmp[64];
		snprintf(tmp, 64, "%s stuck on 24 bytes", mCameraName);
		write_log(tmp);
		CutStream();
		//RestartBackend();
		return 0;
	}

	mPrevSize = buf.st_size;

	if (buf.st_size > 16 * 1024 * 1024) {
		CutStream();
	} else if (now > mStartTime + 120) {
		CutStream();
	}
	setpriority(PRIO_PROCESS, mBackendPID, mBackendPrio);
	return 0;
}
	
int Camera::DeleteOldFiles()
{
	if (mPIDFile) unlink(mPIDFile);
	return 0;
}

int Camera::StartBackend()
{
	char *cameraName;
	char buffer[256];
	char streamName[256];
	char configFile[256];
	int argc;
	char *argv[10];

	write_log("starting tsrep3");
	DeleteOldFiles();
	mSocket->Connect(mSockFile);

	snprintf(buffer, 256, "Video%d", mCameraNo);
	cameraName = Config::GetCharValue(mSection, "CameraName", buffer);
	snprintf(configFile, 256, BACKEND_CONFIG, mCameraNo);
	snprintf(streamName, 256, STREAM_FILE, cameraName);
	snprintf(buffer, 256, "mkdir -p '%s/%s'", STREAM_DIR, cameraName);
	my_system(buffer);

	argc = 0;
	argv[argc++] = "tsrep3";
	argv[argc++] = "-f";
	argv[argc++] = configFile;
	argv[argc++] = "-p";
	argv[argc++] = mPIDFile;
	argv[argc++] = "-n";
	argv[argc++] = mSockFile;
	argv[argc++] = "-o";
	argv[argc++] = streamName;
	argv[argc] = NULL;

	mBackendPID = Execute(argc, argv);
	mRecording = 0;
	mStoppedPending = 0;

	mBackendPrio = getpriority(PRIO_PROCESS, mBackendPID);

	free((void*)cameraName);
	return 0;
}

static void parse_exit_status(int status)
{
	char buf[256];

	if (WIFEXITED(status)) {
		snprintf(buf, 256,
			"tsrep exited normally with exit status %d\n",
			WEXITSTATUS(status));
		write_log(buf);
	}
	if (WIFSIGNALED(status)) {
		if (WTERMSIG(status) != SIGTERM) {
			snprintf(buf, 256,
				"tsrep exited because of unhandled signal %d\n",
				WTERMSIG(status));
			write_log(buf);
		}
	}
#ifdef WCOREDUMP
	if (WCOREDUMP(status)) {
		snprintf(buf, 256,
			"tsrep dumped its core\n");
		write_log(buf);
	}
#endif
}

int Camera::StopBackend()
{
	char buf[256];
	snprintf(buf, 256, "stopping tsrep3 (%d)", mBackendPID);
	write_log(buf);
	if (mBackendPID <= 1) return 0;
	if (!kill(mBackendPID, SIGTERM)) {
		int status;
		waitpid(mBackendPID, &status, 0);
		parse_exit_status(status);
	} else {
		fprintf(stderr, "couldn't send a signal to tsrep3\n");
	}
	mBackendPID = 0;
	return 0;
}

pid_t Camera::Execute(int argc, char *argv[])
{
	int i;
	pid_t pid;

	fprintf(stderr, "executing: ");
	for (i = 0; i < argc; i++) {
		fprintf(stderr, "%s ", argv[i]);
	}
	fprintf(stderr, "\n");
	pid = fork();

	if (pid == 0) {
		if (argv[0][0] == '/') {
			execv(argv[0], argv);
		} else {
			execvp(argv[0], argv);
		}
		fprintf(stderr, "Error executing %s\n", argv[0]);
		exit(-1);
	}
	return pid;
}

int Camera::DumpConfiguration()
{
	char buffer[256];

	if (mBackendPID == 0) return 0;

	if (mAlive) mSocket->Write("RESTART\n");

	PreviewDisplay(":0.0");
	PreviewParent(mWindow);
	PreviewSize(mW, mH);
	PreviewShow();

	RemoteDisplay(":0.1");

	SetCameraName(mCameraName);

	V4LBrightness(mBrightness);
	V4LContrast(mContrast);
	V4LColor(mColor);

	MotionBounds(mBoundsX1, mBoundsY1, mBoundsX2, mBoundsY2);
	SetSensitivity(mSensitivity);
	ShowBounds(mShowBounds);
	ShowMotion(mShowMotion);
	SetMotionRecord(mRecordMotion);
	SetNotification(mNotifyMotion);
	RecordAlways(mRecordAlways);

	V4LBrightness(mBrightness);
	V4LContrast(mContrast);
	V4LColor(mColor);
	SetBitrate(mBitrate);

	if (mRemoteDisplay) RemoteDisplay(mRemoteDisplay);
	if (mRemoteParent) RemoteParent(mRemoteParent);
	if (mRemoteW) RemoteSize(mRemoteW, mRemoteH);
	if (mRemoteVisible) RemoteShow();
	else RemoteHide();

	return 0;
}

int Camera::RenameFile(int flag)
{
	char oldname[256];
	char newname[256];
	char sstr[128];
	char estr[128];
	struct tm tm;
	FILE *f;
	time_t endtime;

	snprintf(oldname, 256, STREAM_FILE, mCameraName);

	if (mPrevSize == -1) {
		fprintf(stderr, "Couldn't start recording, trying again\n");
		unlink(oldname);
		mPrevSize = 0;
		return 0;
	}

	endtime = time(NULL);

	localtime_r(&mStartTime, &tm);
#ifdef USE_NEW_DIR
	char buf[128];
	snprintf(buf, 128, STREAM_DIR"/%s/%02d%s%04d",mCameraName,
			tm.tm_mday, months[tm.tm_mon],tm.tm_year + 1900);
	mkdir(buf, 0777);
#endif

	if (flag) {
		snprintf(sstr, 128, "Interrupt detected at");
	} else {
#ifdef USE_NEW_DIR
		snprintf(sstr, 128, "%02d%02d.%02d",
				tm.tm_hour,tm.tm_min, tm.tm_sec);
#else
		snprintf(sstr, 128, "%04d-%02d-%02d %02d,%02d.%02d",
				tm.tm_year+1900, tm.tm_mon + 1,
				tm.tm_mday, tm.tm_hour,
				tm.tm_min, tm.tm_sec);
#endif
	}

	localtime_r(&endtime, &tm);
#ifdef USE_NEW_DIR
	snprintf(estr, 128, "%02d%02d.%02d",
			tm.tm_hour,tm.tm_min,tm.tm_sec);

	snprintf(newname, 256, "%s/%s_%s.mp4", buf, sstr, estr);
#else
	snprintf(estr, 128, "%04d-%02d-%02d %02d.%02d,%02d",
			tm.tm_year+1900, tm.tm_mon + 1,
			tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec);

	snprintf(newname, 256, STREAM_DIR"/%s/%s %s.mp4",
			mCameraName, sstr, estr);
#endif

	f = fopen(newname, "r");
	if (f) {
		fseek(f, 0, SEEK_END);
		if (ftell(f) == 0) {
			unlink(newname);
		} else {
			char *z = strrchr(newname, '.');
			if (z) *z = '\0';
			strncat(newname, "..mp4", 256);
		}
		fclose(f);
	}
	rename(oldname, newname);
	unlink(oldname);
	return 0;
}

void Camera::TimerCallback(void *data1, void *data2)
{
	Camera *c = (Camera*)data1;

	c->OnTimerCallback();
}
