#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/vfs.h>
#include <sys/stat.h>
#include <time.h>
#include <dirent.h>

#ifndef STREAMS
#define STREAMS "/streams"
#endif

//#define DEBUG

#define ERRLOG_DIR "log"
#define IMAGES_DIR "images"
#define LIVE_FILE "live.mp4"

struct filetree {
        char *path;
	int is_dir;
        time_t creation_time;
        struct filetree *next;
        struct filetree *prev;
};

static struct filetree *ft = NULL;

struct filetree *new_node(char *path, time_t ct, int is_dir)
{
	struct filetree *rv = NULL;

	if (path == NULL) return NULL;

	rv = (struct filetree*)malloc(sizeof(struct filetree));
	if (rv == NULL) return NULL;

	rv->path = strdup(path);
	rv->is_dir = is_dir;
	rv->creation_time = ct;
	rv->next = NULL;
	rv->prev = NULL;
	return rv;
}

int free_node(struct filetree *node)
{
	if (node == NULL) return 0;

	if (node->path) free((void*)node->path);
	free((void*)node);
	return 0;
}

int add_node(struct filetree *node)
{
	struct filetree *f;

	if (node == NULL) return 0;

	if (ft == NULL) {
		ft = node;
		return 0;
	}

	f = ft;

	while (f->next != NULL) {
		f = f->next;
	}

	f->next = node;
	f->next->prev = f;
	f->next->next = NULL;
	return 0;
}

int del_node(struct filetree *node)
{
	struct filetree *f, *n, *p;

	if (ft == NULL) return 0;

	f = ft;
	n = f->next;
	p = f->prev;

	while (f != node) {
		f = f->next;
	}

	n = f->next;
	p = f->prev;
	if (n) n->prev = f->prev;
	if (p) p->next = f->next; 
	else ft = f->next; // special case

	free_node(f);
	return 0;
}

int build_linked_list(char *directory)
{
	DIR *d = NULL;
	struct dirent *entry = NULL;
	struct stat statbuf;

	d = opendir(directory);

	if (d == NULL) return 0;

	while ((entry = readdir(d)) != NULL) {
		char fname[1024];
		if (entry->d_name[0] == '.') continue;
		if (!strcmp(entry->d_name, ERRLOG_DIR)) continue;
		if (!strcmp(entry->d_name, IMAGES_DIR)) continue;
		if (!strcmp(entry->d_name, LIVE_FILE)) continue;
		if (strstr(entry->d_name, "error.log")) continue;
		snprintf(fname, 1024, "%s/%s", directory, entry->d_name);
		stat(fname, &statbuf);
		if (S_ISDIR(statbuf.st_mode)) {
			struct filetree *node;
#ifdef DEBUG
			fprintf(stderr, "d------ %s\n", fname);
#endif
			node = new_node(fname, statbuf.st_mtime, 1);
			add_node(node);
			build_linked_list(fname);
		} else {
			struct filetree *node;
#ifdef DEBUG
			fprintf(stderr, "f------ %s\n", fname);
#endif
			node = new_node(fname, statbuf.st_mtime, 0);
			add_node(node);
		}
	}
	closedir(d);
	return 0;
}

struct filetree *get_oldest_node()
{
	time_t oldest_time;
	struct filetree *node, *rv;

	if (ft == NULL) return NULL;

	node = rv = ft;

	oldest_time = node->creation_time;

	while ((node = node->next) != NULL) {
		if (node->is_dir) continue;
		if (oldest_time > node->creation_time) {
			rv = node;
			oldest_time = node->creation_time;
		}
	}
	return rv;
}

int check_partition_size(char *path, int tolerance)
{
	struct statfs buf;
	char fullpath[256];
	float percent;

	if (path == NULL) return 0;

	snprintf(fullpath, 256, STREAMS "/%s", path);
	statfs(fullpath, &buf);
#ifdef DEBUG
	fprintf(stderr, "checking %s\n", fullpath);
	fprintf(stderr, "free space is %ld\n", buf.f_bfree);
	fprintf(stderr, "total space is %ld\n", buf.f_blocks);
	fprintf(stderr, "percent %0.2f\n",
			((float)(buf.f_bfree)*100.0)/(float)buf.f_blocks);
#endif
	percent = ((float)(buf.f_bfree)*100.0)/(float)buf.f_blocks;
	if (percent < (float)tolerance) {
#ifdef DEBUG
		fprintf(stderr, "not enough free\n");
#endif
		return 1;
	}
	return 0;
}

int delete_old_files(char *path, int tolerance)
{
	struct filetree *node;

#ifdef DEBUG
	fprintf(stderr, "deleting oldest files until we get %d%% free\n",
			tolerance);
#endif
	build_linked_list(STREAMS);

#ifdef DEBUG
	node = ft;
	while (node != NULL) {
		fprintf(stderr, "%s %ld\n", node->path, node->creation_time);
		node = node->next;
	}
#endif

	do {
		node = get_oldest_node();

		if (node) {
#ifdef DEBUG
			fprintf(stderr, "deleting %s %ld\n",
					node->path, node->creation_time);
#endif
			unlink(node->path);
			del_node(node);
		}
	} while ((check_partition_size(path, tolerance)) && (ft != NULL));
	return 0;
}

int delete_empty_dirs()
{
	struct filetree *node;

	node = ft;

	while ((node = node->next) != NULL) {
		if (!node->is_dir) continue;
		rmdir(node->path);
	}
	return 0;
}

#define PIDFILE "/tmp/deleting"

int check_pid()
{
	FILE *f;
	int pid = -1;
	f = fopen(PIDFILE, "r");
#ifdef DEBUG
	fprintf(stderr, "f: %d\n", f);
#endif
	if (f) {
		fscanf(f, "%d", &pid);
		fclose(f);
#ifdef DEBUG
		fprintf(stderr, "pid exists\n");
#endif
	}
	if (pid > -1) {
		char buf[64];
		snprintf (buf, 64, "/proc/%d/cmdline", pid);
		if (!access(buf, F_OK)) {
#ifdef DEBUG
			fprintf(stderr, "delete is running\n");
#endif
			exit(0);
		} else {
#ifdef DEBUG
			fprintf(stderr, "stale lock file\n");
#endif
			unlink(PIDFILE);
		}
	}
	return 0;
}

int open_pid()
{
	FILE *f;
	f = fopen(PIDFILE, "w");
	if (f == NULL) {
#ifdef DEBUG
		fprintf(stderr, "could not open pid file\n");
#endif
		return 0;
	}
	fprintf(f, "%d", getpid());
	fclose(f);
	return 0;
}

int close_pid()
{
	unlink(PIDFILE);
	return 0;
}

int do_delete(char *cameraName, int tolerance)
{
	check_pid();
	open_pid();
	
	nice(20);
	if (cameraName == NULL) return -1;

	if (check_partition_size(cameraName, tolerance)) {
		delete_old_files(cameraName, tolerance);
		delete_empty_dirs();
	} else
#ifdef DEBUG
		fprintf(stderr, "no need to delete old files\n");
#endif
	close_pid();
	return 0;
}
