#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <sys/ioctl.h>

#include "modem.h"
#include "debug.h"

#define ECHO_ON 1
#define TIMEOUT 60

modem_t *modem_new()
{
	modem_t *rv;
	ENTER;

	rv = (modem_t*)malloc(sizeof(modem_t));
	if (rv == NULL) return rv;

	memset(rv, 0, sizeof(modem_t));
	LEAVE;
	return rv;
}

int modem_free(modem_t *modem)
{
	ENTER;
	if (modem == NULL) return 0;

	if (modem->phone_number) {
		free((void*)modem->phone_number);
		modem->phone_number = NULL;
	}
	if (modem->audio_buf) {
		free((void*)modem->audio_buf);
		modem->audio_buf = 0;
	}
	free((void*)modem);
	LEAVE;
	return 0;
}

int modem_open_device(modem_t *modem, char *device)
{
	struct termios pts, sts;

	ENTER;
	if (modem == NULL) return -1;
	if (device == NULL) return -1;

	if (modem->dev_fd) modem_close_device(modem);

	modem->dev_fd = open(device, O_RDWR);
	if (modem->dev_fd <= 0) {
		modem->dev_fd = 0;
		return -1;
	}

	tcgetattr(modem->dev_fd, &pts);
	pts.c_lflag &= ~ICANON;
	pts.c_lflag &= ~(ECHO |ECHOCTL | ECHONL);
	pts.c_cflag |= HUPCL;
	pts.c_cc[VMIN] = 1;
	pts.c_cc[VTIME] = 0;
	pts.c_oflag &= ~ONLCR;
	pts.c_iflag &= ~ICRNL;
	pts.c_cflag |= CRTSCTS;
	pts.c_iflag &= ~(IXON | IXOFF | IXANY);
	cfsetospeed(&pts, B115200);
	cfsetispeed(&pts, B115200);
	tcsetattr(modem->dev_fd, TCSANOW, &pts);
	modem_send_line(modem, "AT&F");
	if (modem_check_response(modem, "OK")) {
		fprintf(stderr, "could not reset the modem\n");
	}
	modem_send_line(modem, "AT L0 M0");
	if (modem_check_response(modem, "OK")) {
		fprintf(stderr, "could not set the speaker volume\n");
	}
#if ECHO_ON
	modem_send_line(modem, "ATE1");
#else
	modem_send_line(modem, "ATE0");
#endif
	modem_check_response(modem, "OK");
	modem_send_line(modem, "AT+FCLASS=8");
	if (modem_check_response(modem, "OK")) {
		fprintf(stderr, "could not set voice mode\n");
		exit(1);
	}
	modem_send_line(modem, "AT+VIP");
	if (modem_check_response(modem, "OK")) {
		fprintf(stderr, "could not init voice params\n");
	}

	modem_send_line(modem, "AT+VSM=141");
	if (modem_check_response(modem, "OK")) {
		fprintf(stderr, "could not set audio mode\n");
		exit(1);
	}
	modem_send_line(modem, "AT+VSD=128,100");
	if (modem_check_response(modem, "OK")) {
		fprintf(stderr, "could not set volume\n");
	}
	modem_send_line(modem, "AT+VSG=80");
	if (modem_check_response(modem, "OK")) {
		fprintf(stderr, "could not set gain\n");
	}
	modem_send_line(modem, "AT+VNH=0");
	if (modem_check_response(modem, "OK")) {
		fprintf(stderr, "could not set auto hangup\n");
	}

	LEAVE;
	return 0;
}

int modem_close_device(modem_t *modem)
{
	if (modem == NULL) return -1;
	if (modem->dev_fd <= 0) return 0;

	ENTER;
	close(modem->dev_fd);
	modem->dev_fd = 0;
	LEAVE;
	return 0;
}

int modem_send_buffer(modem_t *modem, char *buf, unsigned long buf_size)
{
	if (modem == NULL) return -1;
	if (buf == NULL) return -1;
	if (buf_size == 0) return 0;
	if (modem->dev_fd <= 0) return 0;

	write(modem->dev_fd, buf, buf_size);
	return 0;
}

int modem_send_line(modem_t *modem, char *line)
{
	int rv = 0;
	char buf[128];
	ENTER;
	if (modem == NULL) return -1;
	if (line == NULL) return -1;
	
	snprintf(buf, 128, "%s\r\n", line);
	fprintf(stderr, "[modem <--] %s", buf);
	LEAVE;
	rv = modem_send_buffer(modem, buf, strlen(buf));
	if (ECHO_ON) rv = modem_check_response(modem, line);
	return rv;
}

char *modem_get_line(modem_t *modem)
{
	char rv[256];
	char b;
	int i = 0;

	ENTER;
	if (modem == NULL) return NULL;
	if (modem->dev_fd <= 0) return NULL;

	memset(rv, 0, 256);
	alarm(60);
	do {
		int j = 0;
try_again:
		if ((j = read(modem->dev_fd, &b, 1) <= 0)) {
			goto fin;
		}
		if ((i == 0) && ((b == 13) || (b == 10))) goto try_again;
		rv[i] = b;
		if (i >= 256) goto fin;
		i++;
	} while ((b != '\n') && (b != '\0'));

fin:
	alarm(0);

	if (i) {
		rv[i-1] = '\0';
		fprintf(stderr, "[modem -->] %s\n", rv);
		return strdup(rv);
	}
	LEAVE;
	return NULL;
}

int modem_check_response(modem_t *modem, char *resp)
{
	char *buf = NULL;
	int rv = 0;

	ENTER;
	if (modem == NULL) return -1;

	buf = modem_get_line(modem);
	if (buf == NULL) return -1;
	if (resp) {
		if (strncmp(buf, resp, strlen(resp))) rv = -1;
	}
	free((void*)buf);
	LEAVE;
	return rv;
}

int modem_dial(modem_t *modem)
{
	char buf[256];
	int rv = 0;

	ENTER;
	if (modem == NULL) return -1;
	if (modem->phone_number == NULL) return -1;

	if (modem->offhook) modem_hangup(modem);

	modem->offhook = 1;
	snprintf(buf, 256, "ATDT%s;", modem->phone_number);
	modem_send_line(modem, buf);

	rv = modem_check_response(modem, "OK");
	LEAVE;
	return rv;
}

int modem_set_number(modem_t *modem, char *phone_number)
{
	ENTER;
	if (modem == NULL) return -1;
	if (modem->phone_number) {
		free((void*)modem->phone_number);
		modem->phone_number = NULL;
	}
	if (phone_number) {
		modem->phone_number = strdup(phone_number);
	}
	LEAVE;
	return 0;
}

int modem_set_audio(modem_t *modem, char *buf, unsigned long size)
{
	ENTER;
	if (modem == NULL) return -1;
	if (modem->audio_buf) {
		free((void*)modem->audio_buf);
		modem->audio_buf = NULL;
	}

	if ((buf == NULL) || (size == 0)) return 0;

	modem->audio_buf = (char*)malloc(size);
	if (modem->audio_buf == NULL) return -1;

	memcpy(modem->audio_buf, buf, size);
	modem->buf_size = size;
	LEAVE;
	return 0;
}

int modem_hangup(modem_t *modem)
{
	char buf[2] = {0x10, 0x03};
	char *b2;
	int rv = 0;
	
	ENTER;
	if (modem == NULL) return -1;
	if (modem->offhook == 0) return 0;

	modem->offhook = 0;
	modem_send_buffer(modem, buf, 2);
	//modem_check_response(modem, "OK");
	modem_send_line(modem, "AT+FCLASS=0");
	//modem_check_response(modem, "OK");
	modem_send_line(modem, "ATZ");
	//modem_check_response(modem, "OK");
	modem_send_line(modem, "AT&F");
	//modem_check_response(modem, "OK");
	sleep(1);
	LEAVE;
	return rv;
}

int modem_send_message(modem_t *modem)
{
	time_t start = time(NULL);

	ENTER;
	if (modem == NULL) return -1;
	if (modem->phone_number == NULL) return -1;
	if (modem->audio_buf == NULL) return -1;

	modem_open_device(modem, "/dev/ttyS0");
	modem_dial(modem);
	modem_send_line(modem, "AT+VLS=3");
	if (modem_check_response(modem, "OK")) {
		fprintf(stderr, "error 1");
		goto error;
	}
	do {
		char buf[2] = {0x10, 0x03};
		modem_send_line(modem, "AT+VTX");
		//modem_check_response(modem, "WHOCARES");
		modem_send_buffer(modem, modem->audio_buf, modem->buf_size);
		modem_send_buffer(modem, buf, 2);
		sleep(1);
		fprintf(stderr, "sending %lu bytes at %d (%d seconds to go)               \r",
				modem->buf_size, time(NULL), start + TIMEOUT - time(NULL));
	} while (time(NULL) < start + TIMEOUT);

error:
	modem_hangup(modem);
	modem_close_device(modem);
	LEAVE;
	return 0;
}
