#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <sys/shm.h>

#include "memcpy.h"
#include "graphics.h"

pthread_mutex_t Text::mMutex = PTHREAD_MUTEX_INITIALIZER;
char *Text::mText = NULL;

#ifndef toupper
char toupper(char a)
{
	if (a >= 'a' && a <= 'z') a-=32;
	return a;
}
#endif

PreviewWindow::PreviewWindow()
{
	display = NULL;
	gc = (GC)0;
	window = (Window)0;
	parent = (Window)0;
	width = 0;
	height = 0;
	mGotShm = 0;
}

PreviewWindow::~PreviewWindow()
{
}

int PreviewWindow::DestroyImage()
{
	if (mGotShm) {
		if (mShmInfo.shmaddr) {
			XShmDetach(display, &mShmInfo);
			XDestroyImage(mImage);
			shmdt(mShmInfo.shmaddr);
		}
	} else {
		if (mImage) XDestroyImage(mImage);
	}

	return 0;
}

int PreviewWindow::CreateImage(int w, int h)
{
	DestroyImage();
	Screen *screen =  DefaultScreenOfDisplay(display);
	Visual *visual = DefaultVisualOfScreen(screen);
	if (mGotShm) {
		mImage = XShmCreateImage(display, visual,
				24, //DefaultDepthOfScreen(screen),
				ZPixmap, NULL, &mShmInfo,
				w, h);

		if (mImage == NULL) {
			mGotShm = 0;
			return CreateImage(w, h);
		}

		mShmInfo.shmid = shmget(IPC_PRIVATE,
				mImage->bytes_per_line * mImage->height,
				IPC_CREAT | 0777);
		if (mShmInfo.shmid < 0) {
			mGotShm = 0;
			return CreateImage(w, h);
		}

		mShmInfo.shmaddr=mImage->data=(char*)shmat(mShmInfo.shmid,0,0);
		if (mShmInfo.shmaddr == (char*)-1) {
			mGotShm = 0;
			return CreateImage(w, h);
		}

		mShmInfo.readOnly = False;
		XShmAttach(display, &mShmInfo);
		XSync(display, False);
		shmctl( mShmInfo.shmid,IPC_RMID,0 );

	} else {
		mImage = XCreateImage(display, visual,
				DefaultDepthOfScreen(screen),
				ZPixmap, 0, NULL, w, h, 8, 0);
		mImage->data = (char*)malloc(mImage->bytes_per_line * mImage->height);
	}
	fprintf(stderr, "Image Created %d %d\n", mImage->width, mImage->height);
	return 0;
}

int PreviewWindow::open_display(char *d)
{
	if (display != NULL) return 0;
	display = ::XOpenDisplay(d);

	if (display == NULL) {
		fprintf(stderr, "could not connect to %s\n", d);
		return 0;
	}

	if (XShmQueryExtension(display)) {
		mGotShm = 1;
	} else {
		fprintf(stderr, "no XShm extension found\n");
	}
	return 0;
}

int PreviewWindow::close_display()
{
	if (display == NULL) return -1;

	XCloseDisplay(display);
	display = NULL;
	return 0;
}

int PreviewWindow::create_window(Window p)
{
	if (display == NULL) return -1;

	if (p) parent = p;
	else parent = DefaultRootWindow(display);

	window = XCreateSimpleWindow(display, parent, 0, 0, 2, 2,
			0, 0, 0x0);
	XSelectInput(display, window, ButtonPressMask | ButtonReleaseMask);
	CreateImage(384,288);
	XCreateGC(display, window, 0, NULL);
	XSetWindowBackground(display, window, BlackPixel(display, 0));
	return 0;
}

int PreviewWindow::set_parent(Window p)
{
	if (display == NULL) return -1;
	if (window == (Window)0) {
		return create_window(p);
	}

	if (p) parent = p;
	else parent = DefaultRootWindow(display);

	XSynchronize(display, True);
	XReparentWindow(display, window, parent, 0, 0);
	XSynchronize(display, False);
	return 0;
}

int PreviewWindow::set_size(int width, int height)
{
	if ((width == 0) || (height == 0)) return -1;
	if (display == NULL) return -1;
	if (window == (Window)0) return -1;

	XResizeWindow(display, window, width, height);

	CreateImage(width, height);
	return 0;
}

int PreviewWindow::show()
{
	if (display == NULL) return -1;
	if (window == (Window)0) return -1;

	XMapWindow(display, window);
	return 0;
}

int PreviewWindow::hide()
{
	if (display == NULL) return -1;
	if (window == (Window)0) return -1;

	XUnmapWindow(display, window);
	return 0;
}

int PreviewWindow::sync()
{
	if (display == NULL) return -1;

	return 0;
}

int PreviewWindow::get_pending()
{
	XEvent ev;
	int rv = 0;
	if (display == NULL) return 0;

	if (XPending(display)) {
		XNextEvent(display, &ev);
		switch (ev.type) {
			case ButtonPress:
				rv = BUTTON_PRESS | ev.xbutton.button;
				break;
			case ButtonRelease:
				rv = BUTTON_RELEASE | ev.xbutton.button;
				break;
		}
	}

	return rv;
}

int PreviewWindow::ScaleImage(struct ts_frame *f,
		int srcW, int srcH, int dstW, int dstH)
{
	char *src, *dest;
	
	dest = mImage->data;
	src = (char*)f->rgb;
	// Easy case :)
	if ((srcW == dstW) && (srcH == dstH)) {
		memcpy(dest, src, f->fh.width*f->fh.height*4);
		return 1;
	} else {
		int x, y;
		int sx, sy;
		int tx, ty;
		char *p;

		tx = srcW * 65536 / dstW;
		ty = srcH * 65536 / dstH;
		sy = 0;
		for (y = 0; y < dstH; y++) {
			p = src + (sy >> 16) * srcW *4;
			sx = 0;
			for (x = 0; x < dstW; x++) {
				*dest++ = p[(sx>>16)*4];
				*dest++ = p[(sx>>16)*4+1];
				*dest++ = p[(sx>>16)*4+2];
				*dest++ = p[(sx>>16)*4+3];
				sx += tx;
			}
			sy += ty;
		}
		return 1;
	}
	return 0;
}

int PreviewWindow::show_next_frame(struct ts_frame *f)
{
	if (display == NULL) return -1;
	if (window == (Window)0) return -1;
	if (f == NULL) return -1;

	if (f->fh.format != 15) return 0;

	if (mGotShm) {
		if (!gc) gc = XCreateGC(display, window, 0, NULL);

		if (ScaleImage(f, f->fh.width,f->fh.height,
					mImage->width,mImage->height)) {
			XShmPutImage(display, window, gc, mImage, 0, 0, 0, 0,
					mImage->width, mImage->height, False);
			XSync(display, false);
		}
	} else {
		if (!gc) gc = XCreateGC(display, window, 0, NULL);

		if (ScaleImage(f, f->fh.width, f->fh.height,
					mImage->width,mImage->height)) {
			XPutImage(display, window, gc, mImage, 0, 0, 0, 0,
					mImage->width, mImage->height);
			XSync(display, false);
		}
	}

	return 0;
}


struct draw_char {
	char ascii;
	char pix[6][6];
};

struct draw_char draw_table[]= {
	{
		'0',
		{
			{0,0,1,1,1,0},
			{0,1,0,0,1,1},
			{0,1,0,1,0,1},
			{0,1,0,1,0,1},
			{0,1,1,0,0,1},
			{0,0,1,1,1,0}
		}
	},
	{
		'1',	
		{
			{0,0,0,1,0,0},
			{0,0,1,1,0,0},
			{0,0,0,1,0,0},
			{0,0,0,1,0,0},
			{0,0,0,1,0,0},
			{0,0,1,1,1,0}
		}
	},
	{
		'2',
		{
			{0,0,1,1,0,0},
			{0,1,0,0,1,0},
			{0,0,0,0,1,0},
			{0,0,0,1,0,0},
			{0,0,1,0,0,0},
			{0,1,1,1,1,0}
		}
	},
	{
		'3',
		{
			{0,0,1,1,1,0},
			{0,1,0,0,0,1},
			{0,0,0,1,1,0},
			{0,0,0,0,0,1},
			{0,1,0,0,0,1},
			{0,0,1,1,1,0}
		}
	},
	{
		'4',
		{
			{0,0,0,0,1,0},
			{0,0,0,1,1,0},
			{0,0,1,0,1,0},
			{0,1,1,1,1,1},
			{0,0,0,0,1,0},
			{0,0,0,0,1,0}
		}
	},
	{
		'5',
		{
			{0,1,1,1,1,1},
			{0,1,0,0,0,0},
			{0,1,1,1,1,0},
			{0,0,0,0,0,1},
			{0,0,0,0,0,1},
			{0,1,1,1,1,0}
		}
	},
	{
		'6',
		{
			{0,0,1,1,1,1},
			{0,1,0,0,0,0},
			{0,1,1,1,1,0},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,0,1,1,1,0}
		}
	},
	{
		'7',
		{
			{0,1,1,1,1,1},
			{0,0,0,0,0,1},
			{0,0,0,0,1,0},
			{0,0,0,1,0,0},
			{0,0,1,0,0,0},
			{0,0,1,0,0,0}
		}
	},
	{
		'8',
		{
			{0,0,1,1,1,0},
			{0,1,0,0,0,1},
			{0,0,1,1,1,0},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,0,1,1,1,0}
		}
	},
	{
		'9',
		{
			{0,0,1,1,1,0},
			{0,1,0,0,0,1},
			{0,0,1,1,1,1},
			{0,0,0,0,0,1},
			{0,1,0,0,0,1},
			{0,0,1,1,1,0}
		}
	},
	{
		':',
		{
			{0,0,1,1,0,0},
			{0,0,1,1,0,0},
			{0,0,0,0,0,0},
			{0,0,0,0,0,0},
			{0,0,1,1,0,0},
			{0,0,1,1,0,0},
		}
	},
	{
		'/',
		{
			{0,0,0,0,0,1},
			{0,0,0,0,1,0},
			{0,0,0,1,0,0},
			{0,0,1,0,0,0},
			{0,1,0,0,0,0},
			{1,0,0,0,0,0},
		}
	},
	{
		'-',
		{
			{0,0,0,0,0,0},
			{0,0,0,0,0,0},
			{0,1,1,1,1,1},
			{0,0,0,0,0,0},
			{0,0,0,0,0,0},
			{0,0,0,0,0,0}
		}
	},
		{
		'A',
		{
			{0,0,1,1,1,0},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,1,1,1,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1}
		}
	},
	{
		'B',
		{
			{0,1,1,1,1,0},
			{0,1,0,0,0,1},
			{0,1,1,1,1,0},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,1,1,1,0}
		}
	},
	{
		'C',
		{
			{0,0,1,1,1,0},
			{0,1,0,0,0,1},
			{0,1,0,0,0,0},
			{0,1,0,0,0,0},
			{0,1,0,0,0,1},
			{0,0,1,1,1,0}
		}
	},
	{
		'D',
		{
			{0,1,1,1,1,0},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,1,1,1,0}
		}
	},
	{
		'E',
		{
			{0,1,1,1,1,1},
			{0,1,0,0,0,0},
			{0,1,1,1,1,0},
			{0,1,0,0,0,0},
			{0,1,0,0,0,0},
			{0,1,1,1,1,1}
		}
	},
	{
		'F',
		{
			{0,1,1,1,1,1},
			{0,1,0,0,0,0},
			{0,1,1,1,1,0},
			{0,1,0,0,0,0},
			{0,1,0,0,0,0},
			{0,1,0,0,0,0}
		}
	},
	{
		'G',
		{
			{0,0,1,1,1,0},
			{0,1,0,0,0,1},
			{0,1,0,0,0,0},
			{0,1,0,1,1,1},
			{0,1,0,0,0,1},
			{0,0,1,1,1,0}
		}
	},
	{
		'H',
		{
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,1,1,1,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1}
		}
	},
	{
		'I',
		{
			{0,0,1,1,1,0},
			{0,0,0,1,0,0},
			{0,0,0,1,0,0},
			{0,0,0,1,0,0},
			{0,0,0,1,0,0},
			{0,0,1,1,1,0}
		}
	},
	{
		'J',
		{
			{0,0,1,1,1,1},
			{0,0,0,0,1,0},
			{0,0,0,0,1,0},
			{0,0,0,0,1,0},
			{0,1,0,0,1,0},
			{0,0,1,1,0,0}
		}
	},
	{
		'K',
		{
			{0,1,0,0,0,1},
			{0,1,0,0,1,0},
			{0,1,0,1,0,0},
			{0,1,1,1,0,0},
			{0,1,0,0,1,0},
			{0,1,0,0,0,1}
		}
	},
	{
		'L',
		{
			{0,1,0,0,0,0},
			{0,1,0,0,0,0},
			{0,1,0,0,0,0},
			{0,1,0,0,0,0},
			{0,1,0,0,0,0},
			{0,1,1,1,1,0}
		}
	},
	{
		'M',
		{
			{0,1,0,0,0,1},
			{0,1,1,0,1,1},
			{0,1,0,1,0,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1}
		}
	},
	{
		'N',
		{
			{0,1,0,0,0,1},
			{0,1,1,0,0,1},
			{0,1,0,1,0,1},
			{0,1,0,0,1,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1}
		}
	},
	{
		'O',
		{
			{0,0,1,1,1,0},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,0,1,1,1,0}
		}
	},
	{
		'P',
		{
			{0,1,1,1,1,0},
			{0,1,0,0,0,1},
			{0,1,1,1,1,0},
			{0,1,0,0,0,0},
			{0,1,0,0,0,0},
			{0,1,0,0,0,0}
		}
	},
	{
		'Q',
		{
			{0,0,1,1,1,0},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,1,0,1},
			{0,1,0,0,1,1},
			{0,0,1,1,1,1}
		}
	},
	{
		'R',
		{
			{0,1,1,1,1,0},
			{0,1,0,0,0,1},
			{0,1,1,1,1,0},
			{0,1,0,1,0,0},
			{0,1,0,0,1,0},
			{0,1,0,0,0,1}
		}
	},
	{
		'S',
		{
			{0,0,1,1,1,1},
			{0,1,0,0,0,0},
			{0,0,1,1,1,0},
			{0,0,0,0,0,1},
			{0,0,0,0,0,1},
			{0,1,1,1,1,0}
		}
	},
	{
		'T',
		{
			{0,1,1,1,1,1},
			{0,0,0,1,0,0},
			{0,0,0,1,0,0},
			{0,0,0,1,0,0},
			{0,0,0,1,0,0},
			{0,0,0,1,0,0}
		}
	},
	{
		'U',
		{
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,0,1,1,1,1}
		}
	},
	{
		'V',
		{
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,0,1,0,1,0},
			{0,0,0,1,0,0}
		}
	},
	{
		'W',
		{
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,0,0,1},
			{0,1,0,1,0,1},
			{0,1,1,0,1,1},
			{0,1,0,0,0,1}
		}
	},
	{
		'X',
		{
			{0,1,0,0,0,1},
			{0,0,1,0,1,0},
			{0,0,0,1,0,0},
			{0,0,0,1,0,0},
			{0,0,1,0,1,0},
			{0,1,0,0,0,1}
		}
	},
	{
		'Y',
		{
			{0,1,0,0,0,1},
			{0,0,1,0,1,0},
			{0,0,0,1,0,0},
			{0,0,0,1,0,0},
			{0,0,0,1,0,0},
			{0,0,0,1,0,0}
		}
	},
	{
		'Z',
		{
			{0,1,1,1,1,1},
			{0,0,0,0,1,0},
			{0,0,0,1,0,0},
			{0,0,1,0,0,0},
			{0,1,0,0,0,0},
			{0,1,1,1,1,1}
		}
	},
	{
		' ',
		{
			{0,0,0,0,0,0},
			{0,0,0,0,0,0},
			{0,0,0,0,0,0},
			{0,0,0,0,0,0},
			{0,0,0,0,0,0},
			{0,0,0,0,0,0}
		}
	},
	{
		0,
		{
			{0,0,0,0,0,0},
			{0,0,0,0,0,0},
			{0,0,0,0,0,0},
			{0,0,0,0,0,0},
			{0,0,0,0,0,0},
			{0,0,0,0,0,0}
		}
	}
};

int Text::DrawText(char *yuvbuf, int x, int y, int w, char *text)
{
	int i, pos=0, x1, y1;

	if ((mText == NULL) && (text == NULL)) return 0;
	if (text == NULL) text = mText;

	pthread_mutex_lock(&mMutex);
	if (x>w/2) {
		x-=strlen(text)*6;
		if (x<0)
			x=0;
	}
	while (text[pos]) {
		if (x+(pos+1)*6>=w)
			break;
		i=0;
		while (draw_table[i].ascii && draw_table[i].ascii!=toupper(text[pos])) i++;
		for (x1=0; x1<6; x1++) {
			for (y1=0; y1<6; y1++) {
				if (draw_table[i].pix[y1][x1]) {
					yuvbuf[x1+x+pos*6+1+(y1+y)*w]=0;
					yuvbuf[x1+x+pos*6-1+(y1+y)*w]=0;
					yuvbuf[x1+x+pos*6+(y1+y+1)*w]=0;
					yuvbuf[x1+x+pos*6+(y1+y-1)*w]=0;
				}
			}
		}
		for (x1=0; x1<6; x1++) {
			for (y1=0; y1<6; y1++) {
				if (draw_table[i].pix[y1][x1])
					yuvbuf[x1+x+pos*6+(y1+y)*w]=255;
			}
		}
		pos++;
	}
	pthread_mutex_unlock(&mMutex);
	return 0;
}

int Text::SetText(char *text)
{
	pthread_mutex_lock(&mMutex);
	if (mText) {
		free((void*)mText);
		mText = NULL;
	}
	if (text) mText = strdup(text);
	pthread_mutex_unlock(&mMutex);
	return 0;
}
