#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <sched.h>

#include "mp4live.h"
#include "media_flow.h"
#include "graphics.h"
#include "message.h"
#include "socket.h"
#include "motion.h"
#include "yuv2rgb.h"
#include "tsrep_options.h"

#include "memcpy.h"

extern int v4l_pipe;

MessageHandler *pMsg = NULL;
Socket *pSocket = NULL;

int send_message_to_frontend(char *message)
{
	if (pSocket == NULL) {
		fprintf(stderr, "null socket\n");
		return 0;
	}

	return pSocket->write(message, strlen(message));
}

void sig_handler(int sig)
{
	if (sig == SIGHUP) {
		fprintf(stderr, "got sighup\n");
		return;
	}
	pMsg->StopRecording();
}

int grab_frame(struct ts_frame *f)
{
	unsigned long voffset, uoffset;

	if (v4l_pipe == -1) return -1;

	voffset = f->fh.width * f->fh.height;
	uoffset = voffset * 5 / 4;

	read(v4l_pipe, &(f->fh), sizeof(struct frame_header));
	f->data = (unsigned char*)malloc(f->fh.length);
	if (f->data == NULL) {
		f->fh.length = 0;
		return 0;
	}
	memset(f->data, 0, f->fh.length);
	int i = read(v4l_pipe, f->data, f->fh.length);

	if (f->rgb == NULL) {
		if (f->fh.width < 384) {
			fprintf(stderr, "\n\n\n BAD CONFIGURATION:\n");
			fprintf(stderr, "width != 384\n");
			fprintf(stderr, "\n\n\n");
			fprintf(stderr, "killing entire devsec\n");
			system("killall -TERM devsec_watchdog devsec tsrep3");
			exit(1);
		}
		f->rgb = (unsigned char*)malloc(f->fh.width * f->fh.height * 4);
	}

        yuv2rgb((uint8_t*)f->rgb, f->data, f->data + voffset,
                        f->data + uoffset,
                        f->fh.width, f->fh.height,
                        f->fh.width * 4, f->fh.width, f->fh.width / 2);

	return 0;
}

int write_pidfile()
{
	FILE *f;

	if (TsrepOpts::GetPidFile() == NULL) {
		fprintf(stderr, "NULL pidfile, expect problems\n");
		return -1;
	}

	f = fopen(TsrepOpts::GetPidFile(), "w");
	if (f == NULL) {
		fprintf(stderr, "cannot open pid file\n");
		return -1;
	}
	fclose(f);
	return 0;
}

int check_sched_old()
{
	static time_t delta = time(NULL);
	static int frames = 0;
	static sched_param p = {0};
	static int old_prio = 0;

	frames++;
	if (delta != time(NULL)) {
		frames = 0;
		delta = time(NULL);
		if (frames < 12) p.sched_priority++;
		if (frames > 12) p.sched_priority--;
		if (p.sched_priority < 1) p.sched_priority = 1;
		if (p.sched_priority > 99) p.sched_priority = 99;
		if (old_prio != p.sched_priority) {
			old_prio = p.sched_priority;
			sched_setscheduler(0, SCHED_RR, &p);
		}
	}
	return 0;
}

static struct timeval old_time = {0};
static sched_param sp = {0};
static int old_prio = 0;

int check_sched()
{
	struct timeval now;
	long long delta;

	gettimeofday(&now, NULL);

	delta = (now.tv_sec - old_time.tv_sec) * 1000000 + 
		(now.tv_usec - old_time.tv_usec);

	old_time.tv_sec = now.tv_sec;
	old_time.tv_usec = now.tv_usec;

	delta -= 60000;
	delta /= 10000;
	sp.sched_priority += delta;

	if (sp.sched_priority < 1) sp.sched_priority = 1;
	if (sp.sched_priority > 99) sp.sched_priority = 99;
	if (old_prio != sp.sched_priority) {
		old_prio = sp.sched_priority;
		sched_setscheduler(0, SCHED_RR, &sp);
	}
	return 0;
}

int gui_main(int argc, char **argv, CLiveConfig *pConfig)
{
	char buf[1024];
	struct ts_frame f;
	time_t stop_time = 0;

	//probe_fast_memcpy(MEMCPY_SSE);

	signal(SIGHUP, sig_handler);
	memset(&f, 0, sizeof(struct ts_frame));
	f.rgb = 0;
	yuv2rgb_init(32, MODE_RGB);

	CAVMediaFlow *pFlow = new CAVMediaFlow(pConfig);
	pMsg = new MessageHandler(pFlow, pConfig);
	PreviewWindow *pLocalPreview = new PreviewWindow();
	PreviewWindow *pRemotePreview = new PreviewWindow();

	pSocket = new Socket(TsrepOpts::GetSockFile());
	Motion *pMotion = new Motion();

	pMsg->set_objects(pLocalPreview, pRemotePreview, pSocket, pMotion);
	pConfig->m_appAutomatic = true;
	pConfig->SetStringValue(CONFIG_RECORD_MP4_FILE_NAME, TsrepOpts::GetOutput());

	write_pidfile();

	pMotion->sent_motion_event = 0;

	gettimeofday(&old_time, NULL);
	for (;;) {
		check_sched();
		while (pSocket->pending()) {
			pSocket->read(buf, 1024);
			pMsg->handle_message(buf);
		}
		if (!grab_frame(&f)) {
			if (pMotion->detect_motion(f.rgb)) {
				stop_time = 0;
				if (!pMotion->sent_motion_event) {
					pSocket->write("EV_MOTION_START\n", 16);
					pMotion->sent_motion_event = 1;
				}
			} else {
				if (!stop_time) stop_time = time(NULL);
				if (stop_time + 60 < time(NULL)) {
					if (pMotion->sent_motion_event) {
					pSocket->write("EV_MOTION_STOP\n", 15);
					pMotion->sent_motion_event = 0;
					}
				}
			}
			pLocalPreview->show_next_frame(&f);
			pRemotePreview->show_next_frame(&f);
			int c;
			while ((c = pLocalPreview->get_pending()) != 0) {
				char buf[32];
				if (c && BUTTON_PRESS) {
					c = c & ~BUTTON_PRESS;
					snprintf(buf, 32, "EV_MOUSE_DOWN %d\n",
							c);
					pSocket->write(buf, strlen(buf));
				}
				if (c && BUTTON_RELEASE) {
					c = c & ~BUTTON_PRESS;
					snprintf(buf, 32, "EV_MOUSE_UP %d\n",
							c);
					pSocket->write(buf, strlen(buf));
				}
			}
			free((void*)f.data);
		}
	}

	return 0;
}
