#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "graphics.h"
#include "mp4live.h"
#include "media_flow.h"
#include "message.h"
#include "motion.h"

MessageHandler::MessageHandler(CAVMediaFlow *pFlow, CLiveConfig *pConfig)
{
	mFlow = pFlow;
	mConfig = pConfig;
	mLocalPreview = NULL;
	mRemotePreview = NULL;
	mSocket = NULL;
	mMotion = NULL;
	recording = 0;

	mMp4Recorder = new CMp4Recorder();
	mMp4Recorder->SetConfig(mConfig);
	mMp4Recorder->StartThread();

	mRtpTransmitter = NULL;
}

int MessageHandler::set_objects(PreviewWindow *lp, PreviewWindow *rp,
		Socket *s, Motion *m)
{
	mLocalPreview = lp;
	mRemotePreview = rp;
	mSocket = s;
	mMotion = m;
	return 0;
}

int MessageHandler::StartRecording()
{
	if (recording) return 0;

	mFlow->AddSink(mMp4Recorder);
	mMp4Recorder->Start();
	recording = 1;
	mFlow->GenerateKeyFrame();
	return 0;
}

int MessageHandler::StopRecording()
{
	if (mMp4Recorder == NULL) return 0;

	mMp4Recorder->Stop();
	mFlow->RemoveSink(mMp4Recorder);
	recording = 0;
	return 0;
}

int MessageHandler::StartUnicast(char *addr, char *port)
{
	int p = atoi(port);
	if (mRtpTransmitter) StopUnicast();

	if (p == 0) p = mConfig->GetIntegerValue(CONFIG_RTP_VIDEO_DEST_PORT);
	mRtpTransmitter = new CRtpTransmitter(mConfig);
	mRtpTransmitter->StartThread();
	mRtpTransmitter->CreateVideoRtpDestination(0, addr, p, 0);
	mFlow->AddSink(mRtpTransmitter);
	mRtpTransmitter->Start();
	return 0;
}

int MessageHandler::StopUnicast()
{
	if (!mRtpTransmitter) return 0;
	mFlow->RemoveSink(mRtpTransmitter);
	mRtpTransmitter->StopThread();
	delete mRtpTransmitter;
	mRtpTransmitter = NULL;
	return 0;
}

int MessageHandler::handle_v4l_message(char *message)
{
	unsigned int val;

	if (message == NULL) return 0;
	if (message[0] == '\0') return 0;

	if (!strncmp(message, "bright", 6)) {
		val = 32767;
		sscanf(message, "bright=%u", &val);
		val = val * 100 / 65535;
		mConfig->SetIntegerValue(CONFIG_VIDEO_BRIGHTNESS,
				(config_integer_t)val);
	} else if (!strncmp(message, "contrast", 8)) {
		val = 32767;
		sscanf(message, "contrast=%u", &val);
		val = val * 100 / 65535;
		mConfig->SetIntegerValue(CONFIG_VIDEO_CONTRAST,
				(config_integer_t)val);
	} else if (!strncmp(message, "hue", 3)) {
		val = 32767;
		sscanf(message, "hue=%u", &val);
		val = val * 100 / 65535;
		mConfig->SetIntegerValue(CONFIG_VIDEO_HUE,
				(config_integer_t)val);
	} else if (!strncmp(message, "color", 5)) {
		val = 32767;
		sscanf(message, "color=%u", &val);
		val = val * 100 / 65535;
		mConfig->SetIntegerValue(CONFIG_VIDEO_COLOR,
				(config_integer_t)val);
	}
	mFlow->SetPictureControls();
	return 0;
}

int MessageHandler::handle_motion_message(char *message)
{
	int a, b, c, d;
	unsigned int val;

	if (message == NULL) return 0;
	if (message[0] == '\0') return 0;
	if (mMotion == NULL) return 0;

	if (!strncmp(message, "bounds", 6)) {
		a = b = 0;
		c = 384;
		d = 288;
		sscanf(message, "bounds=%dx%dx%dx%d", &a, &b, &c, &d);
		return mMotion->set_bounds(a, b, c, d);
	} else if (!strncmp(message, "showbounds", 10)) {
		val = 0;
		sscanf(message, "showbounds=%u", &val);
		return mMotion->set_showbounds(val);
	} else if (!strncmp(message, "showmotion", 10)) {
		val = 0;
		sscanf(message, "showmotion=%u", &val);
		return mMotion->set_showmotion(val);
	} else if (!strncmp(message, "resetmotion", 11)) {
		mMotion->sent_motion_event = 0;
	}
	return 0;
}

int MessageHandler::handle_preview_message(char *message)
{
	char *p;
	unsigned long l;
	unsigned int a, b;
	PreviewWindow *preview = NULL;
	
	if (message == NULL) return 0;
	if (message[0] == '\0') return 0;

	p = &message[0];
	if (!strncmp(message, "local", 5)) {
		p = &message[6];
		preview = mLocalPreview;
	} else if (!strncmp(message, "remote", 6)) {
		p = &message[7];
		preview = mRemotePreview;
	}

	if (!preview) return 0;

	if (!strncmp(p, "size", 4)) {
		a = 384;
		b = 288;
		sscanf(p, "size=%ux%u", &a, &b);
		preview->set_size(a, b);
	} else if (!strncmp(p, "parent", 6)) {
		l = 0;
		sscanf(p, "parent=%lu", &l);
		if (l) preview->set_parent((Window)l);
	} else if (!strncmp(p, "display", 7)) {
		preview->open_display(&p[8]);
	} else if (!strncmp(p, "show", 4)) {
		preview->show();
	} else if (!strncmp(p, "hide", 4)) {
		preview->hide();
	}
	preview->sync();
	return 0;
}

int MessageHandler::handle_mpeg_message(char *message)
{
	unsigned int val;

	if (message == NULL) return 0;
	if (message[0] == '\0') return 0;

	if (!strncmp(message, "start", 5)) {
		StartRecording();
		mFlow->GenerateKeyFrame();
		return 0;
	} else if (!strncmp(message, "stop", 4)) {
		StopRecording();
		return 0;
	} else if (!strncmp(message, "bitrate", 7)) {
		sscanf(message, "bitrate=%u", &val);
		mConfig->SetIntegerValue(CONFIG_VIDEO_BIT_RATE,
				(config_integer_t)val);
	}
	return 0;
}

int MessageHandler::handle_text_message(char *message)
{
	if (message == NULL) return 0;
	if (!strncmp(message, "text", 4)) {
		Text::SetText(&message[5]);
	}
	return 0;
}

int MessageHandler::handle_restart_message()
{
	mConfig->Update();
	mFlow->Stop();
	mFlow->Start();
	return 0;
}

int MessageHandler::handle_unicast_message(char *message)
{
	char destAddr[64];
	char destPort[32];

	if (message == NULL) return 0;
	switch (sscanf(message, "%s %s", &destAddr, &destPort)) {
		case 2:
			StartUnicast(destAddr, destPort);
			break;
		case 1:
		case 0:
			StopUnicast();
			break;
	}
	return 0;
}

int MessageHandler::handle_message(char *message)
{
	if (message == NULL) return 0;

	if (!strncmp(message, "V4L", 3))
		return handle_v4l_message(&message[4]);
	else if (!strncmp(message, "MOTION", 6))
		return handle_motion_message(&message[7]);
	else if (!strncmp(message, "PREVIEW", 7))
		return handle_preview_message(&message[8]);
	else if (!strncmp(message, "MPEG", 4))
		return handle_mpeg_message(&message[5]);
	else if (!strncmp(message, "TEXT", 4))
		return handle_text_message(&message[5]);
	else if (!strncmp(message, "RESTART", 7))
		return handle_restart_message();
	else if (!strncmp(message, "UNICAST", 7))
		return handle_unicast_message(&message[8]);
	return 0;
}
