#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "motion.h"
#include "memcpy.h"

#define XBLOCKS 64 // Must evenly divide width
#define YBLOCKS 96 // Must evenly divide height

Motion::Motion()
{
	mDepth = 4;
	mVideoBuf[0] = (unsigned char*)malloc(384 * 288 * mDepth);
	mVideoBuf[1] = (unsigned char*)malloc(384 * 288 * mDepth);
	mBlocks[0] = (rgb24_t*)malloc(sizeof(rgb24_t)*XBLOCKS*YBLOCKS);
	mBlocks[1] = (rgb24_t*)malloc(sizeof(rgb24_t)*XBLOCKS*YBLOCKS);
	mActiveBuf = 0;
	mFirst = 1;
	mX1 = 0;
	mY1 = 0;
	mX2 = 383;
	mY2 = 287;
	mSensitivity = 160;
	mTicks = 50;
}

Motion::~Motion()
{
	if (mVideoBuf[0]) free((void*)mVideoBuf[0]);
	if (mVideoBuf[1]) free((void*)mVideoBuf[1]);
	if (mBlocks[0]) free((void*)mBlocks[0]);
	if (mBlocks[1]) free((void*)mBlocks[1]);
}

int Motion::set_sensitivity(unsigned int sensitivity)
{
	mSensitivity = sensitivity;
	return 0;
}

int Motion::set_bounds(unsigned long x1, unsigned long y1,
		unsigned long x2, unsigned long y2)
{
	mX1 = x1;
	mY1 = y1;
	mX2 = x2;
	mY2 = y2;
	return 0;
}

int Motion::set_showmotion(unsigned char showmotion)
{
	mShowMotion = showmotion;
	return 0;
}

int Motion::set_showbounds(unsigned char showbounds)
{
	mShowBounds = showbounds;
	return 0;
}

int Motion::detect_motion(unsigned char *buffer)
{
	int rv = 0;
	unsigned long x1, y1, x2, y2;

	memcpy(mVideoBuf[mActiveBuf], buffer, 384 * 288 * mDepth);
	process_image();
	mActiveBuf = 1 - mActiveBuf;
	if (mTicks) {
		mTicks--;
		return 0;
	}
	if (mFirst) {
		mFirst = 0;
	} else {
		rv = compare_blocks(&x1, &y1, &x2, &y2);
		if (rv) {
			if (mShowMotion) draw_box(buffer, x1, y1, x2, y2);
		} else if (mShowBounds) {
			draw_box(buffer, mX1, mY1, mX2, mY2);
		}
	}
	return rv;
}

int Motion::process_image()
{
	unsigned char *p;
	int x, y, m, n;
	int blk_wid;
	int blk_hgt;
	int x0, y0, x1, y1;
	int s;
	unsigned int R, G, B;
	int r_offset, g_offset;
	int startx, starty, endx, endy;
	rgb24_t *r;

	blk_wid = 384 / XBLOCKS;
	blk_hgt = 288 / YBLOCKS;

	s = blk_wid * blk_hgt;

	r_offset = 2;
	g_offset = 1;

	memset(mBlocks[mActiveBuf], 0, sizeof(rgb24_t)*XBLOCKS*YBLOCKS);

	startx = mX1 / blk_wid + 1;
	starty = mY1 / blk_hgt + 1;
	endx = mX2 / blk_wid;
	endy = mY2 / blk_hgt;

	r = mBlocks[mActiveBuf];
	r += (starty-1) * XBLOCKS + (startx-1);
	for (n = starty; n <= endy; n++) {
		for (m = startx; m <= endx; m++) {
			x0 = (m-1) * blk_wid;
			y0 = (n-1) * blk_hgt;
			x1 = blk_wid * m - 1;
			y1 = blk_hgt * n - 1;
			R = G = B = 0;
			for (y = y0; y <= y1; y++) {
				for (x = x0; x <= x1; x++) {
					p = mVideoBuf[mActiveBuf] +
						(y * 384 + x) * mDepth;
					R += *(p + r_offset);
					G += *(p + g_offset);
					B += *p;
				}
			}
			r->red = R / s;
			r->green = G / s;
			r->blue = B / s;
			r++;
		}
		r += (startx-1) + (XBLOCKS - endx);
	}
	return 0;
}

int Motion::compare_blocks(unsigned long *x1, unsigned long *y1,
		unsigned long *x2, unsigned long *y2)
{
	unsigned long x, y;
	int i = 0;
	int blk_wid;
	int blk_hgt;
	unsigned int color_diff;
	rgb24_t *a, *b;
	int rv = 0;

	a = mBlocks[0];
	b = mBlocks[1];

	*x1 = 384;
	*y1 = 288;
	*x2 = *y2 = 0;

	blk_wid = 384 / XBLOCKS;
	blk_hgt = 288 / YBLOCKS;

	for (y = 0; y < YBLOCKS; y++) {
		for (x = 0; x < XBLOCKS; x++) {
			color_diff = abs(a[i].red - b[i].red) +
				abs(a[i].green - b[i].green) +
				abs(a[i].blue - b[i].blue);
			if (color_diff > mSensitivity) {
				rv = 1;
				if (mShowMotion) {
					add_point(x1, y1, x2, y2, x * blk_wid,
							y * blk_hgt);
				} else goto got_motion;
			}
			i++;
		}
	}
got_motion:
	return rv;
}

int Motion::add_point(unsigned long *x1, unsigned long *y1,
		unsigned long *x2, unsigned long *y2,
		unsigned long x, unsigned long y)
{
	if (*x1 > x) *x1 = x;
	if (*y1 > y) *y1 = y;
	if (*x2 < x) *x2 = x;
	if (*y2 < y) *y2 = y;
	return 0;
}

int Motion::draw_box(unsigned char *buffer,
		unsigned long x1, unsigned long y1,
		unsigned long x2, unsigned long y2)
{
	unsigned long i;

	for (i = x1; i < x2; i++) {
		buffer[(i + y1 * 384) * mDepth] = 0xff;
		buffer[(i + y1 * 384) * mDepth + 1] = 0xff;
		buffer[(i + y1 * 384) * mDepth + 2] = 0xff;
		buffer[(i + y1 * 384) * mDepth + 3] = 0xff;
		buffer[(i + y2 * 384) * mDepth] = 0xff;
		buffer[(i + y2 * 384) * mDepth + 1] = 0xff;
		buffer[(i + y2 * 384) * mDepth + 2] = 0xff;
		buffer[(i + y2 * 384) * mDepth + 3] = 0xff;
	}

	for (i = y1; i < y2; i++) {
		buffer[(x1 + i * 384) * mDepth] = 0xff;
		buffer[(x1 + i * 384) * mDepth + 1] = 0xff;
		buffer[(x1 + i * 384) * mDepth + 2] = 0xff;
		buffer[(x1 + i * 384) * mDepth + 3] = 0xff;
		buffer[(x2 + i * 384) * mDepth] = 0xff;
		buffer[(x2 + i * 384) * mDepth + 1] = 0xff;
		buffer[(x2 + i * 384) * mDepth + 2] = 0xff;
		buffer[(x2 + i * 384) * mDepth + 3] = 0xff;
	}

	return 0;
}
