/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is MPEG4IP.
 * 
 * The Initial Developer of the Original Code is Cisco Systems Inc.
 * Portions created by Cisco Systems Inc. are
 * Copyright (C) Cisco Systems Inc. 2000, 2001.  All Rights Reserved.
 * 
 * Contributor(s): 
 *		Dave Mackie		dmackie@cisco.com
 *		Bill May 		wmay@cisco.com
 */

#define __TESTING 1

#include "mp4live.h"
#include <unistd.h>

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "video_v4l_source.h"
#include "video_util_rgb.h"
#include "video_util_filter.h"
#include "tsrep_com.h"
#include "gui/graphics.h"

int v4l_pipe = -1;

//#define DEBUG_TIMESTAMPS 1
int CV4LVideoSource::ThreadMain(void) 
{
	setpriority(PRIO_PROCESS, 0, 20);
	if (v4l_pipe == -1) v4l_pipe = tsrep_com_open_pipe();
	while (true) {
		int rc;

		if (m_source) {
			rc = SDL_SemTryWait(m_myMsgQueueSemaphore);
		} else {
			rc = SDL_SemWait(m_myMsgQueueSemaphore);
		}

		// semaphore error
		if (rc == -1) {
			break;
		} 

		// message pending
		if (rc == 0) {
			CMsg* pMsg = m_myMsgQueue.get_message();
		
			if (pMsg != NULL) {
				switch (pMsg->get_value()) {
				case MSG_NODE_STOP_THREAD:
					DoStopCapture();	// ensure things get cleaned up
					delete pMsg;
					return 0;
				case MSG_NODE_START:
				case MSG_SOURCE_START_VIDEO:
					DoStartCapture();
					break;
				case MSG_NODE_STOP:
					DoStopCapture();
					break;
				case MSG_SOURCE_KEY_FRAME:
					DoGenerateKeyFrame();
					break;
				}

				delete pMsg;
			}
		}

		if (m_source) {
			try {
				ProcessVideo();
			}
			catch (...) {
				DoStopCapture();
				break;
			}
		}
	}
	tsrep_com_close_pipe();

	return -1;
}

void CV4LVideoSource::DoStartCapture(void)
{
	if (m_source) {
		return;
	}
	if (!Init()) {
		return;
	}

	m_source = true;
}

void CV4LVideoSource::DoStopCapture(void)
{
	if (!m_source) {
		return;
	}

	DoStopVideo();

	ReleaseDevice();

	m_source = false;
}

bool CV4LVideoSource::Init(void)
{
	if (!InitDevice()) {
		return false;
	}

	m_pConfig->CalculateVideoFrameSize();

	InitVideo(
		(m_pConfig->m_videoNeedRgbToYuv ?
			RGBVIDEOFRAME :
			YUVVIDEOFRAME),
		true);

	SetVideoSrcSize(
		m_pConfig->GetIntegerValue(CONFIG_VIDEO_RAW_WIDTH),
		m_pConfig->GetIntegerValue(CONFIG_VIDEO_RAW_HEIGHT),
		m_pConfig->GetIntegerValue(CONFIG_VIDEO_RAW_WIDTH),
		true);

	m_maxPasses = (u_int8_t)(m_videoSrcFrameRate + 0.5);

	return true;
}

bool CV4LVideoSource::InitDevice(void)
{
	char *active = "ACTIVE\n";
	int rc;
	const char* deviceName = m_pConfig->GetStringValue(CONFIG_VIDEO_SOURCE_NAME);
	u_int16_t format;

	do {
		m_videoDevice = open(deviceName, O_RDWR);
		if (m_videoDevice < 0) {
			error_message("Failed to open %s: %s", 
				deviceName, strerror(errno));
			if (errno == ENODEV) {
				m_videoDevice = 0;
				error_message("No device\n");
			} else {
				error_message("Sleeping for 1 second");
				sleep(1);
			}
		}
	} while (m_videoDevice < 0);

	send_message_to_frontend(active);

	// get device capabilities
	struct video_capability videoCapability;
	if (m_videoDevice == 0) {
		videoCapability.name[0] = 'T';
		videoCapability.name[1] = 'e';
		videoCapability.name[2] = 's';
		videoCapability.name[3] = 't';
		videoCapability.name[4] = '\0';
		videoCapability.type = 235;
		videoCapability.channels = 4;
		videoCapability.audios = 3;
		videoCapability.maxwidth = 924;
		videoCapability.maxheight = 576;
		videoCapability.minwidth = 48;
		videoCapability.minheight = 32;
	}
	else {
		rc = ioctl(m_videoDevice, VIDIOCGCAP, &videoCapability);
		if (rc < 0) {
			error_message("Failed to get video capabilities for %s",
				deviceName);
			goto failure;
		}
	}
#ifdef DEBUG
	fprintf(stderr, "videoCapability:\n");
	fprintf(stderr, "name: %s\n", videoCapability.name);
	fprintf(stderr, "type: %d\n", videoCapability.type);
	fprintf(stderr, "channels: %d\n", videoCapability.channels);
	fprintf(stderr, "audios: %d\n", videoCapability.audios);
	fprintf(stderr, "maxwidth: %d\n", videoCapability.maxwidth);
	fprintf(stderr, "maxheight: %d\n", videoCapability.maxheight);
	fprintf(stderr, "minwidth: %d\n", videoCapability.minwidth);
	fprintf(stderr, "minheight: %d\n", videoCapability.minheight);
	fprintf(stderr, "--------\n");
#endif

	if (!(videoCapability.type & VID_TYPE_CAPTURE)) {
		error_message("Device %s is not capable of video capture!",
			deviceName);
		goto failure;
	}

	// N.B. "channel" here is really an input source
	struct video_channel videoChannel;
	videoChannel.channel = m_pConfig->GetIntegerValue(CONFIG_VIDEO_INPUT);
	if (m_videoDevice == 0) {
		videoChannel.name[0] = 'T';
		videoChannel.name[1] = 'e';
		videoChannel.name[2] = 's';
		videoChannel.name[3] = 't';
		videoChannel.name[4] = '\0';
		videoChannel.tuners = 0;
		videoChannel.flags = 2;
		videoChannel.type = 2;
		videoChannel.norm = 0;
		m_videoSrcFrameRate = VIDEO_PAL_FRAME_RATE;
	} else {
		rc = ioctl(m_videoDevice, VIDIOCGCHAN, &videoChannel);
		if (rc < 0) {
			error_message("Failed to get video channel info for %s",
				deviceName);
			goto failure;
		}
		// select video input and signal type
		switch(m_pConfig->GetIntegerValue(CONFIG_VIDEO_SIGNAL)) {
		case VIDEO_SIGNAL_PAL:
		  videoChannel.norm = VIDEO_MODE_PAL;
		  break;
		case VIDEO_SIGNAL_NTSC:
		  videoChannel.norm = VIDEO_MODE_NTSC;
		  break;
		case VIDEO_SIGNAL_SECAM:
		  videoChannel.norm = VIDEO_MODE_SECAM;
		  break;
		}
		rc = ioctl(m_videoDevice, VIDIOCSCHAN, &videoChannel);
		if (rc < 0) {
			error_message("Failed to set video channel info for %s",
				deviceName);
			goto failure;
		}
		if (m_pConfig->GetIntegerValue(CONFIG_VIDEO_SIGNAL) == VIDEO_SIGNAL_NTSC) {
			m_videoSrcFrameRate = VIDEO_NTSC_FRAME_RATE;
		} else {
			m_videoSrcFrameRate = VIDEO_PAL_FRAME_RATE;
		}

		// input source has a TV tuner
		if (videoChannel.flags & VIDEO_VC_TUNER) {
			struct video_tuner videoTuner;

			// get tuner info
			if ((int32_t)m_pConfig->GetIntegerValue(CONFIG_VIDEO_TUNER) == -1) {
				m_pConfig->SetIntegerValue(CONFIG_VIDEO_TUNER, 0);
			}
			videoTuner.tuner = m_pConfig->GetIntegerValue(CONFIG_VIDEO_TUNER);
			rc = ioctl(m_videoDevice, VIDIOCGTUNER, &videoTuner);
			if (rc < 0) {
				error_message("Failed to get video tuner info for %s",
					deviceName);
			}
			
			// set tuner and signal type
			videoTuner.mode = m_pConfig->GetIntegerValue(CONFIG_VIDEO_SIGNAL);
			rc = ioctl(m_videoDevice, VIDIOCSTUNER, &videoTuner);
			if (rc < 0) {
				error_message("Failed to set video tuner info for %s",
					deviceName);
			}

			// tune in the desired frequency (channel)
			struct CHANLISTS* pChannelList = chanlists;
			struct CHANLIST* pChannel = pChannelList[
				m_pConfig->GetIntegerValue(CONFIG_VIDEO_CHANNEL_LIST_INDEX)].list;
			unsigned long videoFrequencyKHz = pChannel[
				m_pConfig->GetIntegerValue(CONFIG_VIDEO_CHANNEL_INDEX)].freq;
			unsigned long videoFrequencyTuner =
				((videoFrequencyKHz / 1000) << 4) 
				| ((videoFrequencyKHz % 1000) >> 6);

			rc = ioctl(m_videoDevice, VIDIOCSFREQ, &videoFrequencyTuner);
			if (rc < 0) {
				error_message("Failed to set video tuner frequency for %s",
					deviceName);
			}
		}
	}
	fprintf(stderr, "videoChannel\n");
	fprintf(stderr, "channel: %d\n", videoChannel.channel);
	fprintf(stderr, "name: %s\n", videoChannel.name);
	fprintf(stderr, "tuners: %d\n", videoChannel.tuners);
	fprintf(stderr, "flags: %d\n", videoChannel.flags);
	fprintf(stderr, "type: %d\n", videoChannel.type);
	fprintf(stderr, "norm: %d\n", videoChannel.norm);
	fprintf(stderr, "--------\n");

	// get info on video capture buffers 
	if (m_videoDevice == 0) {
		m_videoMbuf.size = 8519680;
		m_videoMbuf.frames = 4;
		m_videoMbuf.offsets[0] = 0;
		m_videoMbuf.offsets[1] = 2129920;
		m_videoMbuf.offsets[2] = 4259840;
		m_videoMbuf.offsets[3] = 6389760;
	} else {
		rc = ioctl(m_videoDevice, VIDIOCGMBUF, &m_videoMbuf);
		if (rc < 0) {
			error_message("Failed to get video capture info for %s", 
				deviceName);
			goto failure;
		}
		fprintf(stderr, "videoMbuf:\n");
		fprintf(stderr, "size: %d\n", m_videoMbuf.size);
		fprintf(stderr, "frames: %d\n", m_videoMbuf.frames);
		for (int i = 0; i < VIDEO_MAX_FRAME; i++) {
			if (m_videoMbuf.frames > i)
				fprintf(stderr, "offsets[%d]: %d\n", i, m_videoMbuf.offsets[i]);
		}
	}

        // map the video capture buffers
	if (m_videoDevice == 0) {
		m_videoMap = malloc(m_videoMbuf.size);
	} else {
		m_videoMap = mmap(0, m_videoMbuf.size, 
			PROT_READ | PROT_WRITE, MAP_SHARED, m_videoDevice, 0);
		if (m_videoMap == MAP_FAILED) {
			error_message("Failed to map video capture memory for %s", 
				deviceName);
			goto failure;
		}
	}

	// allocate enough frame maps
	m_videoFrameMap = (struct video_mmap*)
		malloc(m_videoMbuf.frames * sizeof(struct video_mmap));
	if (m_videoFrameMap == NULL) {
		error_message("Failed to allocate enough memory"); 
		goto failure;
	}

	m_videoFrameMapFrame = (uint64_t *)malloc(m_videoMbuf.frames * sizeof(uint64_t));
	m_videoFrameMapTimestamp = (uint64_t *)malloc(m_videoMbuf.frames * sizeof(Timestamp));
	m_captureHead = 0;
	m_encodeHead = -1;
	format = VIDEO_PALETTE_YUV420P;

	m_videoFrames = 0;
       m_videoSrcFrameDuration = 
		(Duration)(((float)TimestampTicks / m_videoSrcFrameRate) + 0.5);
       m_cacheTimestamp = true;
       if (m_pConfig->GetBoolValue(CONFIG_V4L_CACHE_TIMESTAMP) == false) {
	 m_videoMbuf.frames = 2;
	 m_cacheTimestamp = false;
       }
       width = m_pConfig->GetIntegerValue(CONFIG_VIDEO_RAW_WIDTH);
       height = m_pConfig->GetIntegerValue(CONFIG_VIDEO_RAW_HEIGHT);
       
       if (strncasecmp(m_pConfig->GetStringValue(CONFIG_VIDEO_FILTER),
		       VIDEO_FILTER_DECIMATE,
		       strlen(VIDEO_FILTER_DECIMATE)) == 0) {
	 uint32_t max_width, max_height;
	 switch (m_pConfig->GetIntegerValue(CONFIG_VIDEO_SIGNAL)) {
	 case VIDEO_SIGNAL_NTSC:
	   max_width = 720;
	   max_height = 480;
	   break;
	 case VIDEO_SIGNAL_PAL:
	 case VIDEO_SIGNAL_SECAM:
	 default:
	   max_width = 768;
	   max_height = 576;
	   break;
	 }
	 if (max_width < width * 2 || max_height < height * 2) {
	   error_message("Decimate filter choosen with too large video size - max %ux%u",
			 max_width / 2, max_height / 2);
	 } else {
	   m_decimate_filter = true;
	   width *= 2;
	   height *= 2;
	 }
       }

	for (int i = 0; i < m_videoMbuf.frames; i++) {
		// initialize frame map
		m_videoFrameMap[i].frame = i;
		m_videoFrameMap[i].width = width;
		m_videoFrameMap[i].height = height;
		m_videoFrameMap[i].format = format;

		// give frame to the video capture device
		if (m_videoDevice) {
			rc = ioctl(m_videoDevice, VIDIOCMCAPTURE, &m_videoFrameMap[i]);
			if (rc < 0) {
				// try RGB24 palette instead
				if (i == 0 && format == VIDEO_PALETTE_YUV420P) {
					format = VIDEO_PALETTE_RGB24;
					i--;
					continue;
				} 

				error_message("Failed to allocate video capture buffer for %s - frame %d max %d rc %d errno %d format %d", 
					deviceName,
					      i, m_videoMbuf.frames, 
					      rc, errno, format);
				goto failure;
			}
		}
		if (i == 0) {
		  m_videoCaptureStartTimestamp = GetTimestamp();
		}
		m_lastVideoFrameMapFrameLoaded = m_videoFrameMapFrame[i] = i;
		m_lastVideoFrameMapTimestampLoaded = 
		  m_videoFrameMapTimestamp[i] = 
		  CalculateVideoTimestampFromFrames(i);
	}

	if (format == VIDEO_PALETTE_RGB24) {
		m_pConfig->m_videoNeedRgbToYuv = true;
	}

	SetPictureControls();

	if (videoCapability.audios) {
		SetVideoAudioMute(false);
	}

	return true;

failure:
	free(m_videoFrameMap);
	if (m_videoMap) {
		munmap(m_videoMap, m_videoMbuf.size);
		m_videoMap = NULL;
	}
	close(m_videoDevice);
	m_videoDevice = -1;
	return false;
}

void CV4LVideoSource::ReleaseDevice()
{
	SetVideoAudioMute(true);

	// release device resources
	if (m_videoDevice == 0) {
		free((void*)m_videoMap);
	} else {
		munmap(m_videoMap, m_videoMbuf.size);
	}
	m_videoMap = NULL;

	if (m_videoDevice) close(m_videoDevice);
	m_videoDevice = -1;
}
	
void CV4LVideoSource::SetVideoAudioMute(bool mute)
{
	if (!m_pConfig->m_videoCapabilities
	  || !m_pConfig->m_videoCapabilities->m_hasAudio) {
		return;
	}

	int rc;
	struct video_audio videoAudio;

	if (m_videoDevice == 0) return;

	rc = ioctl(m_videoDevice, VIDIOCGAUDIO, &videoAudio);

	if (rc == 0 && (videoAudio.flags & VIDEO_AUDIO_MUTABLE)) {
		if (mute) {
			videoAudio.flags |= VIDEO_AUDIO_MUTE;
		} else {
			videoAudio.flags &= ~VIDEO_AUDIO_MUTE;
		}

		rc = ioctl(m_videoDevice, VIDIOCSAUDIO, &videoAudio);

		if (rc < 0) {
			debug_message("Can't set video audio for %s",
				m_pConfig->m_videoCapabilities->m_deviceName);
		}
	}
}

bool CV4LVideoSource::SetPictureControls()
{
	if (m_videoDevice == -1) {
		return false;
	}

	if (m_videoDevice == 0) return true;

	struct video_picture videoPicture;
	int rc;

	rc = ioctl(m_videoDevice, VIDIOCGPICT, &videoPicture);

	if (rc < 0) {
		return false;
	}

	videoPicture.brightness = (u_int16_t)
		((m_pConfig->GetIntegerValue(CONFIG_VIDEO_BRIGHTNESS) * 0xFFFF) / 100);
	videoPicture.hue = (u_int16_t)
		((m_pConfig->GetIntegerValue(CONFIG_VIDEO_HUE) * 0xFFFF) / 100);
	videoPicture.colour = (u_int16_t)
		((m_pConfig->GetIntegerValue(CONFIG_VIDEO_COLOR) * 0xFFFF) / 100);
	videoPicture.contrast = (u_int16_t)
		((m_pConfig->GetIntegerValue(CONFIG_VIDEO_CONTRAST) * 0xFFFF) / 100);

	rc = ioctl(m_videoDevice, VIDIOCSPICT, &videoPicture);

	if (rc < 0) {
		return false;
	}

	return true;
}

void CV4LVideoSource::TsrepWriteText(struct video_mmap *map, int cf)
{
	char buf[128];
	struct tm tm;
	time_t tt;
	unsigned char *yuvdata;
	static unsigned int width = 0;

	yuvdata = (unsigned char*)m_videoMap + m_videoMbuf.offsets[cf];

	tt = time(NULL);
	localtime_r(&tt, &tm);
	snprintf(buf, 128, "%02d/%02d/%04d %02d:%02d:%02d",
			tm.tm_mday, tm.tm_mon+1,tm.tm_year+1900,
			tm.tm_hour,tm.tm_min,tm.tm_sec);
	if (width == 0) 
		width = m_pConfig->GetIntegerValue(CONFIG_VIDEO_RAW_WIDTH);
	Text::DrawText((char*)yuvdata, width - 14, 4, width, buf);

	Text::DrawText((char*)yuvdata, 4, 4, width);
}

void CV4LVideoSource::SendTsrepFrame(struct video_mmap *map, int cf)
{
	struct ts_frame f;

	f.fh.frame_no = map->frame;
	f.fh.format = map->format;
	f.fh.width = map->width;
	f.fh.height = map->height;
	if (map->format == VIDEO_PALETTE_YUV420P)
		f.fh.length = map->width * map->height * 3 / 2;
	else if (map->format == VIDEO_PALETTE_RGB24)
		f.fh.length = map->width * map->height * 3;

	f.data = (unsigned char*)m_videoMap + m_videoMbuf.offsets[cf];

	tsrep_com_send_frame(&f);
}

int8_t CV4LVideoSource::AcquireFrame(Timestamp &frameTimestamp)
{
	int rc = 1;
	int nv = 0;
	if (m_videoDevice != 0) {
		rc = ioctl(m_videoDevice, VIDIOCSYNC, &m_videoFrameMap[m_captureHead]);
		if (rc == 0) {
			unsigned char *p;
			int x;
			p=(unsigned char*)m_videoMap+m_videoMbuf.offsets[m_captureHead];
			for (x = 0; x < 388; x++) {
				int i = x;
				if (p[i] != 0x23) goto got_video;
				i = i / 2 + m_videoSrcWidth*m_videoSrcHeight;
				if (p[i] != 0xd4) goto got_video;
				i += m_videoSrcWidth*m_videoSrcHeight/4;
				if (p[i] != 0x72) goto got_video;
			}
			rc = 1;
			nv = 1;
		}
	}
got_video:
	if (rc != 0) {
		char *p = (char*)m_videoMap+m_videoMbuf.offsets[m_captureHead];
		char buf[32];
		static unsigned int i = 0;
		static char *no_device = NULL;
		static char *no_video = NULL;
		FILE *f;

		if (no_device == NULL) {
			f = fopen(DTK_PIXMAPDIR "/no_device.yuv","r");
			if (f == NULL) goto img_err;
			no_device = (char*)malloc(m_videoSrcWidth * m_videoSrcHeight * 3 / 2);
			if (!no_device) goto img_err;
			fread(no_device,
					m_videoSrcWidth*m_videoSrcHeight*3/2,
					1, f);
			fclose(f);
			f=fopen(DTK_PIXMAPDIR "/no_signal.yuv","r");
			if (f == NULL) goto img_err;
			no_video = (char*)malloc(m_videoSrcWidth * m_videoSrcHeight * 3 / 2);
			if (!no_video) goto img_err;
			fread(no_video,
					m_videoSrcWidth*m_videoSrcHeight*3/2,
					1, f);
			fclose(f);
		}
		memcpy(p, (nv ? no_video : no_device), m_videoSrcWidth * m_videoSrcHeight * 3 / 2);
img_err:
		if (nv == 0) {
			usleep(1000000/25);
		
			snprintf(buf, 32, "Frame %d", i);

			Text::DrawText((char*)p, m_videoSrcWidth - 20, m_videoSrcHeight - 10, m_videoSrcWidth, buf);
			i++;
		}
	}

	if (m_cacheTimestamp)
	  frameTimestamp = m_videoFrameMapTimestamp[m_captureHead];
	else
	  frameTimestamp = GetTimestamp();

	int8_t capturedFrame = m_captureHead;
	TsrepWriteText(&m_videoFrameMap[capturedFrame], capturedFrame);
	SendTsrepFrame(&m_videoFrameMap[capturedFrame], capturedFrame);
	m_captureHead = (m_captureHead + 1) % m_videoMbuf.frames;

	return capturedFrame;
}
bool CV4LVideoSource::ReleaseFrame(int8_t frameNumber)
{
  Timestamp calc = GetTimestamp();

  if (calc > m_videoSrcFrameDuration + m_lastVideoFrameMapTimestampLoaded) {
#ifdef DEBUG_TIMESTAMPS
    debug_message("video frame delay past end of buffer - time is "U64" should be "U64,
		  calc,
		  m_videoSrcFrameDuration + m_lastVideoFrameMapTimestampLoaded);
#endif
    m_videoCaptureStartTimestamp = calc;
    m_videoFrameMapFrame[frameNumber] = 0;
    m_videoFrameMapTimestamp[frameNumber] = calc;
  } else {
    m_videoFrameMapFrame[frameNumber] = m_lastVideoFrameMapFrameLoaded + 1;
    m_videoFrameMapTimestamp[frameNumber] =
      CalculateVideoTimestampFromFrames(m_videoFrameMapFrame[frameNumber]);
  }

  m_lastVideoFrameMapFrameLoaded = m_videoFrameMapFrame[frameNumber];
  m_lastVideoFrameMapTimestampLoaded = m_videoFrameMapTimestamp[frameNumber];
  if (m_videoDevice) {
	  return (ioctl(m_videoDevice, VIDIOCMCAPTURE, 
			&m_videoFrameMap[frameNumber]) == 0);
  }
  return true;
}

void CV4LVideoSource::ProcessVideo(void)
{
	// for efficiency, process ~1 second before returning to check for commands
  Timestamp frameTimestamp;
	for (int pass = 0; pass < m_maxPasses; pass++) {

		// get next frame from video capture device
		m_encodeHead = AcquireFrame(frameTimestamp);
		if (m_encodeHead == -1) {
			continue;
		}


		u_int8_t* mallocedYuvImage = NULL;
		u_int8_t* pY;
		u_int8_t* pU;
		u_int8_t* pV;

		// perform colorspace conversion if necessary
		if (m_videoSrcType == RGBVIDEOFRAME) {
			mallocedYuvImage = (u_int8_t*)Malloc(m_videoSrcYUVSize);

			pY = mallocedYuvImage;
			pU = pY + m_videoSrcYSize;
			pV = pU + m_videoSrcUVSize,

			RGB2YUV(
				m_videoSrcWidth,
				m_videoSrcHeight,
				(u_int8_t*)m_videoMap + m_videoMbuf.offsets[m_encodeHead],
				pY,
				pU,
				pV,
				1);
		} else {
			pY = (u_int8_t*)m_videoMap + m_videoMbuf.offsets[m_encodeHead];
			pU = pY + m_videoSrcYSize;
			pV = pU + m_videoSrcUVSize;
		}

		if (m_decimate_filter) {
		  video_filter_decimate(pY,
					m_videoSrcWidth,
					m_videoSrcHeight);
		}
		ProcessVideoYUVFrame(
			pY, 
			pU, 
			pV,
			m_videoSrcWidth,
			m_videoSrcWidth >> 1,
			frameTimestamp);

		// release video frame buffer back to video capture device
		if (ReleaseFrame(m_encodeHead)) {
			m_encodeHead = (m_encodeHead + 1) % m_videoMbuf.frames;
		} else {
			debug_message("Couldn't release capture buffer!");
		}
		if (mallocedYuvImage != NULL) {
		  free(mallocedYuvImage);
		}
	}
}

bool CV4LVideoSource::InitialVideoProbe(CLiveConfig* pConfig)
{
	static char* devices[] = {
		"/dev/video", 
		"/dev/video0", 
		"/dev/video1", 
		"/dev/video2", 
		"/dev/video3"
	};
	const char* deviceName = pConfig->GetStringValue(CONFIG_VIDEO_SOURCE_NAME);
	CVideoCapabilities* pVideoCaps;

	// first try the device we're configured with
	pVideoCaps = new CVideoCapabilities(deviceName);

	if (pVideoCaps->IsValid()) {
		pConfig->m_videoCapabilities = pVideoCaps;
		return true;
	}

	delete pVideoCaps;

	return false;

	// no luck, go searching
	for (u_int32_t i = 0; i < sizeof(devices) / sizeof(char*); i++) {

		// don't waste time trying something that's already failed
		if (!strcmp(devices[i], deviceName)) {
			continue;
		} 

		pVideoCaps = new CVideoCapabilities(devices[i]);

		if (pVideoCaps->IsValid()) {
			pConfig->SetStringValue(CONFIG_VIDEO_SOURCE_NAME, devices[i]);
			pConfig->m_videoCapabilities = pVideoCaps;
			return true;
		}
		
		delete pVideoCaps;
	}

	return false;
}

bool CVideoCapabilities::ProbeDevice()
{
	int rc;

	int videoDevice = open(m_deviceName, O_RDWR);
	if (videoDevice < 0) {
		return false;
	}
	m_canOpen = true;

	// get device capabilities
	struct video_capability videoCapability;
	rc = ioctl(videoDevice, VIDIOCGCAP, &videoCapability);
	if (rc < 0) {
		debug_message("Failed to get video capabilities for %s", m_deviceName);
		m_canCapture = false;
		close(videoDevice);
		return false;
	}

	if (!(videoCapability.type & VID_TYPE_CAPTURE)) {
		debug_message("Device %s is not capable of video capture!", 
			m_deviceName);
		m_canCapture = false;
		close(videoDevice);
		return false;
	}
	m_canCapture = true;

	m_driverName = strdup(videoCapability.name);
	m_numInputs = videoCapability.channels;

	m_minWidth = videoCapability.minwidth;
	m_minHeight = videoCapability.minheight;
	m_maxWidth = videoCapability.maxwidth;
	m_maxHeight = videoCapability.maxheight;

	m_hasAudio = videoCapability.audios;

	m_inputNames = (char**)malloc(m_numInputs * sizeof(char*));
	memset(m_inputNames, 0, m_numInputs * sizeof(char*));

	m_inputSignalTypes = (u_int8_t*)malloc(m_numInputs * sizeof(u_int8_t));
	memset(m_inputSignalTypes, 0, m_numInputs * sizeof(u_int8_t));

	m_inputHasTuners = (bool*)malloc(m_numInputs * sizeof(bool));
	memset(m_inputHasTuners, 0, m_numInputs * sizeof(bool));

	m_inputTunerSignalTypes = (u_int8_t*)malloc(m_numInputs * sizeof(u_int8_t));
	memset(m_inputTunerSignalTypes, 0, m_numInputs * sizeof(u_int8_t));


	for (int i = 0; i < m_numInputs; i++) {
		// N.B. "channel" here is really an input source
		struct video_channel videoChannel;
		videoChannel.channel = i;
		rc = ioctl(videoDevice, VIDIOCGCHAN, &videoChannel);
		if (rc < 0) {
			debug_message("Failed to get video channel info for %s:%u",
				m_deviceName, i);
			continue;
		}
		m_inputNames[i] = strdup(videoChannel.name);
		m_inputSignalTypes[i] = videoChannel.norm;

		if (videoChannel.flags & VIDEO_VC_TUNER) {
			// ignore videoChannel.tuners for now
			// current bt drivers only support 1 tuner per input port

			struct video_tuner videoTuner;
			videoTuner.tuner = 0;
			rc = ioctl(videoDevice, VIDIOCGTUNER, &videoTuner);
			if (rc < 0) {
				debug_message("Failed to get video tuner info for %s:%u",
					m_deviceName, i);
				continue;
			}
				
			m_inputHasTuners[i] = true;
			m_inputTunerSignalTypes[i] = videoTuner.flags & 0x7;
		}
	}

	close(videoDevice);
	return true;
}

