#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#define XK_MISCELLANY
#define XK_LATIN1
#include <X11/keysymdef.h>

#include "dtk_object.h"
#include "dtk_basewidget.h"

Display *BaseWidget::mDisplay = NULL;
char *BaseWidget::mDisplayStr = NULL;
BaseWidget *BaseWidget::sFocusedWidget = NULL;

BaseWidget::BaseWidget(BaseWidget *parent, int x, int y, int w, int h)
	: Object()
{
	ENTER;
	Window p;
	XKeyboardControl values;

	if (mDisplay == NULL) {
		mDisplay = XOpenDisplay(mDisplayStr);
	}

	if (mDisplay == NULL) {
		fprintf(stderr, "Couldn't connect to display %s\n",mDisplayStr);
		exit(1);
	}
	
	mDoubleBuffer = 1;
	mOpacity = -1.0;
	mCanHaveFocus = 0;
	mBackground = 0;
	mParent = parent;
	mDrawFrame = 1;
	mVisible = false;
	mTextSize = 12;

	mX = x;
	mY = y;
	mW = (w < 1 ? 1 : w);
	mH = (h < 1 ? 1 : h);
	mEnabled = 1;

	p = DefaultRootWindow(mDisplay);
	if (mParent) p = mParent->GetXWindow();


	values.bell_percent = 100;
	values.bell_pitch = 400;
	values.bell_duration = 64;

	XChangeKeyboardControl(mDisplay,
			KBBellPercent | KBBellPitch | KBBellDuration,
			&values);

	mWindow = XCreateSimpleWindow(mDisplay, p, mX, mY, mW, mH, 0, 0,
			0xc0c0c0);
	mDoubleBufferPixmap = XCreatePixmap(mDisplay, mWindow, mW, mH,
			DefaultDepth(mDisplay, 0));
	mDBGC = XCreateGC(mDisplay, mDoubleBufferPixmap, 0, NULL);

	if (parent == NULL) {
		xa_DeleteWindow = XInternAtom(mDisplay,
				"WM_DELETE_WINDOW", False);
		XSetWMProtocols(mDisplay, mWindow, &xa_DeleteWindow, 1);
	}

	SetEventMask(ButtonPressMask | ButtonReleaseMask | ExposureMask |
			KeyPressMask | KeyReleaseMask);

	mGC = XCreateGC(mDisplay, mWindow, 0, NULL);

	if (mParent) {
		mParent->AddChild(this);
	}
	mPixmap = 0;
	mObjectType = OBJTYPE_WIDGET;
	mText = NULL;
	mTextX = 0;
	mTextY = 0;
	mCallback = NULL;

	mFont = XLoadQueryFont(mDisplay,
			"-*-courier-medium-r-*-*-12-*-*-*-*-*-*-*");
	SetColor(0xc0c0, 0xc0c0, 0xc0c0);
}

BaseWidget::~BaseWidget()
{
	ENTER;

	if (sFocusedWidget == this)
		sFocusedWidget = NULL;

	if (mParent) {
		mParent->RemoveChild(this);
	}
}

int BaseWidget::Destroy()
{
	ENTER;
	if (mGC) XFreeGC(mDisplay, mGC);
	if (mWindow) XDestroyWindow(mDisplay, mWindow);
	if (mDoubleBufferPixmap) XFreePixmap(mDisplay, mDoubleBufferPixmap);
	mDoubleBufferPixmap = 0;
	if (mDBGC) XFreeGC(mDisplay, mDBGC);
	mDBGC = NULL;

	mGC = NULL;
	mWindow = 0;
	mParent = NULL;
	mDirty = 0;
	return 0;
}

int BaseWidget::SetColor(unsigned int r, unsigned int g, unsigned int b)
{
	XColor bg_color;

	if (mWindow == 0) return 0;

	if (mBackground) {
		XFreeColors(mDisplay, DefaultColormap(mDisplay, 0),
					&mBackground, 1, 0);
		mBackground = 0;
	}

	bg_color.pixel = (unsigned long)this;
	bg_color.red = r;
	bg_color.green = g;
	bg_color.blue = b;
	
	if (!XAllocColor(mDisplay, DefaultColormap(mDisplay, 0), &bg_color));

	SetColor(bg_color.pixel);
	return 0;
}

int BaseWidget::SetColor(unsigned long background)
{
	XGCValues gcv;
	mBackground = background;
	mDirty = 1;

	if (!mDBGC) return 0;
	gcv.foreground = background;
	XChangeGC(mDisplay, mDBGC, GCForeground, &gcv);
	XSetWindowBackground(mDisplay, mWindow, background);
	XClearWindow(mDisplay, mWindow);

	BaseWidget *o = (BaseWidget*)mFirstChild;

	while (o != NULL) {
		if (o->GetObjectType() == OBJTYPE_WIDGET)
			o->SetColor(background);
		o = (BaseWidget*)o->GetNextChild();
	}
	return 0;
}

int BaseWidget::SetForegroundColor(unsigned int r,
		unsigned int g, unsigned int b)
{
	XColor fg_color;

	if (mWindow == 0) return 0;

	fg_color.pixel = (unsigned long)this;
	fg_color.red = r;
	fg_color.green = g;
	fg_color.blue = b;
	
	if (!XAllocColor(mDisplay, DefaultColormap(mDisplay, 0), &fg_color));

	SetForegroundColor(fg_color.pixel);
	return 0;
}

int BaseWidget::SetForegroundColor(unsigned long foreground)
{
	XGCValues gcv;

	if (!mGC) return 0;
	mForeground = gcv.foreground = foreground;
	
	XChangeGC(mDisplay, mGC, GCForeground, &gcv);
	BaseWidget *o = (BaseWidget*)mFirstChild;

	while (o != NULL) {
		if (o->GetObjectType() == OBJTYPE_WIDGET)
			o->SetForegroundColor(foreground);
		o = (BaseWidget*)o->GetNextChild();
	}
	return 0;
}

int BaseWidget::SetResizable(int flag)
{
	XSizeHints size_hints;

	size_hints.min_width = size_hints.max_width = mW;
	size_hints.min_height = size_hints.max_height = mH;
	size_hints.flags = PMinSize | PMaxSize;
	XSetWMNormalHints(mDisplay, mWindow, &size_hints);

	return 0;
}

int BaseWidget::Enable(int flag)
{
	mEnabled = flag;
	return 0;
}

int BaseWidget::SetTitle(char *title)
{
	XTextProperty xtp;

	XStringListToTextProperty(&title, 1, &xtp);
	XSetWMName(mDisplay, mWindow, &xtp);
	return 0;
}

int BaseWidget::SetText(char *text, int x, int y)
{
	mDirty = 1;
	if (mText) {
		free((void*)mText);
		mText = NULL;
	}
	if (text == NULL) return 0;
	mText = strdup(text);
	if (x > -1) mTextX = x;
	if (y > -1) mTextY = y;
	return 0;
}

int BaseWidget::SetTextSize(int size)
{
	char fontdesc[128];

	mTextSize = size;
	snprintf(fontdesc, 128, "-*-courier-medium-r-*-*-%d-*-*-*-*-*-*-*",
			size);

	if (mFont) XFree(mFont);
	mFont = XLoadQueryFont(mDisplay, fontdesc);
	return 0;
}

int BaseWidget::SetOpacity(float opacity)
{
	mOpacity = opacity;
	unsigned int newval = (unsigned int) (mOpacity * OPAQUE);

	if (mOpacity == -1.0) return 0;
	if (newval == OPAQUE)
		XDeleteProperty(mDisplay, mWindow,
				XInternAtom(mDisplay, OPACITY, False));
	else
		XChangeProperty(mDisplay, mWindow,
				XInternAtom(mDisplay, OPACITY, False),
				XA_CARDINAL, 32, PropModeReplace,
				(unsigned char*)&newval, 1L);
	XSync(mDisplay, False);
	return 0;
}

int BaseWidget::SetPixmap(Pixmap p)
{
	mPixmap = p;
	if (mPixmap) mDoubleBuffer = 0;
	else mDoubleBuffer = 1;
	mDirty = 1;
	return 0;
}

int BaseWidget::Draw()
{
	ENTER;
	if (mWindow == 0) return 0;

	if (mDirty == 0) return 0;

	mDirty = 0;

	if (mDoubleBuffer) {
		XFillRectangle(mDisplay, mDoubleBufferPixmap, mDBGC,
				0, 0, mW, mH);
	} else {
		if (mPixmap)
			XSetWindowBackgroundPixmap(mDisplay, mWindow, mPixmap);
		else
			XSetWindowBackground(mDisplay, mWindow, mBackground);
		XClearWindow(mDisplay, mWindow);
	}

	if (mDrawFrame) {
		if (mDoubleBuffer) {
			XDrawLine(mDisplay, mDoubleBufferPixmap, mGC, 0, 0,
					mW-1, 0);
			XDrawLine(mDisplay, mDoubleBufferPixmap, mGC, mW-1, 0,
					mW-1, mH-1);
			XDrawLine(mDisplay, mDoubleBufferPixmap, mGC, mW-1,
					mH-1, 0, mH-1);
			XDrawLine(mDisplay, mDoubleBufferPixmap, mGC, 0, mH-1,
					0, 0);
		} else {
			XDrawLine(mDisplay, mWindow, mGC, 0, 0, mW-1, 0);
			XDrawLine(mDisplay, mWindow, mGC, mW-1, 0, mW-1, mH-1);
			XDrawLine(mDisplay, mWindow, mGC, mW-1, mH-1, 0, mH-1);
			XDrawLine(mDisplay, mWindow, mGC, 0, mH-1, 0, 0);
		}
	}

	if (mText) {
		XTextItem textitem;

		textitem.chars = mText;
		textitem.nchars = strlen(mText);
		textitem.delta = 1;
		textitem.font = (mFont ? mFont->fid : None);

		if (mDoubleBuffer) {
			XDrawText(mDisplay, mDoubleBufferPixmap, mGC,
					mTextX, mTextY, &textitem, 1);
		} else {
			XDrawText(mDisplay, mWindow, mGC, mTextX,
					mTextY, &textitem, 1);
		}
	}
	if (mDoubleBuffer) {
		XCopyArea(mDisplay, mDoubleBufferPixmap, mWindow,
				mGC, 0, 0, mW, mH, 0, 0);
	}
	return 0;
}

int BaseWidget::Show()
{
	XEvent event;

	ENTER;
	if (mWindow == 0) return 0;

	if (mVisible) return 0;
	XMapWindow(mDisplay, mWindow);
	XSync(mDisplay, False);
	if (mOpacity > 0) SetOpacity(mOpacity);
	mVisible = true;
	return 0;
}

int BaseWidget::Hide()
{
	ENTER;
	if (mWindow == 0) return 0;

	XUnmapWindow(mDisplay, mWindow);
	mVisible = false;
	return 0;
}

int BaseWidget::Move(int x, int y)
{
	ENTER;
	if (x > -1) mX = x;
	if (y > -1) mY = y;

	return ReallyMove(mX, mY);
}

int BaseWidget::Resize(int w, int h)
{
	ENTER;
	if (w > -1) mW = (w < 1 ? 1 : w);
	if (h > -1) mH = (h < 1 ? 1 : h);

	return ReallyResize(mW, mH);
}

int BaseWidget::ReallyMove(int x, int y)
{
	mX = x;
	mY = y;
	XMoveWindow(mDisplay, mWindow, mX, mY);
	return 0;
}

int BaseWidget::ReallyResize(int w, int h)
{
	mW = w;
	mH = h;
	XResizeWindow(mDisplay, mWindow, mW, mH);
	if (mDoubleBufferPixmap) XFreePixmap(mDisplay, mDoubleBufferPixmap);
	mDoubleBufferPixmap = XCreatePixmap(mDisplay, mWindow, mW, mH, DefaultDepth(mDisplay, 0));
	return 0;
}

int BaseWidget::GetSize(int *x, int *y, int *w, int *h)
{
	if (x) *x = mX;
	if (y) *y = mY;
	if (w) *w = mW;
	if (h) *h = mH;
	return 0;
}

int BaseWidget::Raise()
{
	XRaiseWindow(mDisplay, mWindow);
	return 0;
}

int BaseWidget::MoveResize(int x, int y, int w, int h)
{
	ENTER;
	if (x > -1) mX = x;
	if (y > -1) mY = y;
	if (w > -1) mW = (w < 1 ? 1 : w);
	if (h > -1) mH = (h < 1 ? 1 : h);

	XMoveResizeWindow(mDisplay, mWindow, mX, mY, mW, mH);
	return 0;
}

int BaseWidget::DrawFrame(int flag)
{
	mDrawFrame = flag;
	mDirty = 1;
	return 0;
}

#ifndef MWM_HINTS_DECORATIONS
#define MWM_HINTS_DECORATIONS   (1L << 1)
#endif
typedef struct {
	unsigned long flags;
	unsigned long functions;
	unsigned long decorations;
	         long input_mode;
	unsigned long status;
} MWMHints;

int BaseWidget::WMIgnore(bool yes)
{
	XSetWindowAttributes attr;
	attr.override_redirect = yes;
	//XChangeWindowAttributes(mDisplay, mWindow, CWOverrideRedirect, &attr);
	return 0;
}

int BaseWidget::WMDecorations(bool yes)
{
	Atom WM_HINTS;
	MWMHints hints = {0};
	if (mWindow == 0) return -1;

	WM_HINTS = XInternAtom(mDisplay, "_MOTIF_WM_HINTS", True);
	if (WM_HINTS == None) {
		fprintf(stderr, "can not remove borders\n");
		return -1;
	}

	hints.flags = MWM_HINTS_DECORATIONS;
	hints.decorations = (yes ? 1 : 0);
	XChangeProperty(mDisplay, mWindow, WM_HINTS, WM_HINTS, 32,
			PropModeReplace, (unsigned char*)&hints,
			sizeof(hints)/4);
	return 0;
}

int BaseWidget::IsVisible()
{
	BaseWidget *widget;

	widget = this;
	while (widget != NULL) {
		if (widget->mVisible == 0) return 0;
		widget = widget->mParent;
	}
	return 1;
}

int BaseWidget::EventHandler(TEvent *ev)
{
	int rv = 0;

	ENTER;

	rv = Object::EventHandler(ev);

	if ((sFocusedWidget == NULL) && mCanHaveFocus && IsVisible()) {
		if (mCanHaveFocus) SetFocus();
		return 1;
	}
	if (ev == NULL) return 0;

	if (ev->type == NOEVENT) return 0;

	if (ev->type == XEVENT) {

		switch (ev->xev->type) {
			case ConfigureNotify:
				XWindowAttributes attr;
				XGetWindowAttributes(mDisplay, mWindow, &attr);
				mX = attr.x; mY = attr.y;
				mW = attr.width; mH = attr.height;
				return 1;
			case KeyPress:
				if (sFocusedWidget == this) {
					if (this->HandleKeyPress(ev->xev))
						return 1;
				}
				break;
			case KeyRelease:
				if (sFocusedWidget == this) {
					if (this->HandleKeyPress(ev->xev))
						return 1;
				}
				break;
			case ButtonPress:
				if (ev->xev->xbutton.window != mWindow)
					return 0;
				if (mCallback)
					mCallback(this, ev, mData1, mData2);
				if (mCanHaveFocus) SetFocus();
				break;
			case ButtonRelease:
				if (ev->xev->xbutton.window != mWindow)
					return 0;
				if (mCallback)
					mCallback(this, ev, mData1, mData2);
				break;
			case MotionNotify:
				if (ev->xev->xmotion.window != mWindow)
					return 0;
				if (mCallback)
					mCallback(this, ev, mData1, mData2);
				break;
			case Expose:
				if (ev->xev->xexpose.window != mWindow)
					return 0;
				mDirty = 1;
				break;
			case ClientMessage:
				if (ev->xev->xclient.window != mWindow)
					return 0;
				if (ev->xev->xclient.data.l[0] = xa_DeleteWindow) {
					if (mCallback) {
						TEvent e;
						DTKEvent d;
						d.type = DTK_WINDOW_DESTROY;
						d.destroy.object = (void*)this;
						e.type = DTKEVENT;
						e.dev = &d;
						mCallback(this, &e, mData1,
								mData2);
					}
				}
				break;
		}
	}
	if (mDirty) Draw();
	return rv;
}

int BaseWidget::SetEventMask(long mask)
{
	if (mWindow == 0) return 0;

	XSelectInput(mDisplay, mWindow, mask |
			StructureNotifyMask |
			ButtonPressMask |
			ButtonReleaseMask |
			KeyPressMask |
			KeyReleaseMask |
			ExposureMask );
	return 0;
}

int BaseWidget::LostFocus()
{
	return 0;
}

int BaseWidget::SetFocus()
{
	if (sFocusedWidget) sFocusedWidget->LostFocus();
	sFocusedWidget = this;
	return 0;
}

BaseWidget *BaseWidget::GetFocusedWidget()
{
	return sFocusedWidget;
}

BaseWidget *BaseWidget::GetTopLevelWidget()
{
	BaseWidget *rv;

	rv = this;

	while (rv->GetParentWidget() != NULL) {
		rv = rv->GetParentWidget();
	}
	return rv;
}

BaseWidget *BaseWidget::GetParentWidget()
{
	return mParent;
}

Window BaseWidget::GetXWindow()
{
	ENTER;
	return mWindow;
}

Display *BaseWidget::GetXDisplay()
{
	ENTER;
	return mDisplay;
}

int BaseWidget::SetDisplayStr(char *d)
{
	if (mDisplayStr) {
		free((void*)mDisplayStr);
		mDisplayStr = NULL;
	}
	mDisplayStr = (d ? strdup(d) : NULL);
	return 0;
}

Display *BaseWidget::XConnect()
{
	mDisplay = XOpenDisplay(mDisplayStr);
	return mDisplay;
}

int BaseWidget::GetTextWidth(char *text)
{
	int direction, ascent, descent;
	XCharStruct overall;

	if (text == NULL) return 0;

	XTextExtents(mFont, text, strlen(text),
			&direction, &ascent, &descent, &overall);
	return overall.width;
}

int BaseWidget::SetCallback(void *data1, void*data2,
		void(*cb)(Object*, TEvent*, void*, void*))
{
	mCallback = cb;
	mData1 = data1;
	mData2 = data2;
	return 0;
}

int BaseWidget::HandleKeyPress(XEvent *xev)
{
        char c;
        KeySym keysym;

	if (xev == NULL) return 0;
	if (xev->type == KeyRelease) return 0;
        if (XLookupString(&(xev->xkey), &c, 1, &keysym, NULL) != NoSymbol) {
                if (keysym == XK_Tab) {
			if (sFocusedWidget) sFocusedWidget->LostFocus();
                        sFocusedWidget = NULL;
			return 0;
                }
	}
	if (mCallback) {
		TEvent e;
		e.type = XEVENT;
		e.xev = xev;
		mCallback(this, &e, mData1, mData2);
	}
	return 0;
}

