#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <X11/X.h>
#include <X11/Xlib.h>

#include "dtk_button.h"
#include "dtk_skin.h"

Button::Button(Widget *parent, int x, int y, int w, int h)
	: Widget(parent, x, y, w, h)
{
	ENTER;

	mDrawStipple = 0;
	mNormal = 0;
	mHover = 0;
	mPressed = 0;
	mState = STATE_NORMAL;
	mToggleMode = 0;
	mToggled = 0;
	mButtonPressCount = 0;
	SetEventMask(ButtonPressMask | ButtonReleaseMask |
			PointerMotionMask |
			EnterWindowMask | LeaveWindowMask |
			ExposureMask);
	if ((mW == 80) && (mH == 32)) {
		SetPixmap(Skin::GetSkinPixmap(BUTTON_NORMAL), STATE_NORMAL);
		SetPixmap(Skin::GetSkinPixmap(BUTTON_HOVER), STATE_HOVER);
		SetPixmap(Skin::GetSkinPixmap(BUTTON_PRESSED), STATE_PRESSED);
	}
	mCanHaveFocus = 0;
	SetColor(0x0c0c0, 0x0c0c0, 0x7f7f, STATE_NORMAL);
	SetColor(0x0c0c0, 0x0c0c0, 0x7f7f, STATE_HOVER);
	SetColor(0x0c0c0, 0x7f7f, 0x7f7f, STATE_PRESSED);
	SetForegroundColor(0x0, 0x0, 0x0, STATE_NORMAL);
	SetForegroundColor(0x0, 0x0, 0x0, STATE_HOVER);
	SetForegroundColor(0x0, 0x0, 0x0, STATE_PRESSED);
}

Button::~Button()
{
	ENTER;
}

int Button::Destroy()
{
	Widget::Destroy();
	ENTER;
	return 0;
}

int Button::LostFocus()
{
	if (mToggleMode && mToggled) mState = STATE_PRESSED;
	else mState = STATE_NORMAL;
	mDirty = 1;
}

int Button::SetToggleMode(int flag)
{
	if (!flag) mToggleMode = 0;
	else mToggleMode = 1;
	mDirty = 1;
	return 0;
}

int Button::SetToggle(int flag)
{
	mToggled = flag;
	if (mToggled == 0) mState = STATE_NORMAL;
	mDirty = 1;
	return 0;
}

int Button::EventHandler(TEvent *ev)
{
	int rv = 0;
	ENTER;

	if (ev == NULL) return 0;
	if (ev->type == NOEVENT) return 0;

	if (ev->type == XEVENT) {
		switch (ev->xev->type) {
			case KeyRelease:
				if (sFocusedWidget == this) {
					if (HandleKeyRelease(ev->xev))
						return 1;
				}
				break;
			case ButtonPress:
				if (ev->xev->xbutton.window != mWindow)
					return 0;
				mState = STATE_PRESSED;
				mButtonPressCount++;
				if (mButtonPressCount == 2) return 1;
				if (mToggleMode) {
					mToggled = 1 - mToggled;
				}
				mDirty = 1;
				rv = 1;
				break;
			case ButtonRelease:
				if (ev->xev->xbutton.window != mWindow)
					return 0;
				mButtonPressCount--;
				mState = STATE_HOVER;
				mDirty = 1;
				rv = 1;
				break;
			case EnterNotify:
				if (ev->xev->xcrossing.window != mWindow)
					return 0;
				mState = STATE_HOVER;
				mDirty = 1;
				rv = 1;
				break;
			case LeaveNotify:
				if (ev->xev->xcrossing.window != mWindow)
					return 0;
				mState = STATE_NORMAL;
				mDirty = 1;
				rv = 1;
				break;
		}
	}
	rv = Widget::EventHandler(ev);

	if ((sFocusedWidget == this) && (mState == STATE_NORMAL)) {
		mState = STATE_HOVER;
		mDirty = 1;
	}

	if (mDirty) Draw();

	return rv;
}

int Button::Draw()
{
	Pixmap p = 0;
	XTextItem textitem;
	int rv;

	ENTER;
	if (mWindow == 0) return 0;
	if (mDirty == 0) return 0;

	if (mToggleMode && mToggled) mState = STATE_PRESSED;

	switch (mState) {
		case STATE_NORMAL:
			SetColor(mNormalBackground);
			SetForegroundColor(mNormalForeground);
			mPixmap = mNormal;
			break;
		case STATE_HOVER:
			SetColor(mHoverBackground);
			SetForegroundColor(mHoverForeground);
			mPixmap = mHover;
			break;
		case STATE_PRESSED:
			SetColor(mPressedBackground);
			SetForegroundColor(mPressedForeground);
			mPixmap = mPressed;
			break;
	}

	rv = Widget::Draw();
	if ((sFocusedWidget == this) && (mDrawStipple)) {
		XSetLineAttributes(mDisplay, mGC, 1, LineOnOffDash,
				CapRound, JoinRound);
		XDrawLine(mDisplay, mWindow, mGC, 5, 5, mW - 7, 5);
		XDrawLine(mDisplay, mWindow, mGC, mW - 7, 5, mW - 7, mH - 7);
		XDrawLine(mDisplay, mWindow, mGC, 5, mH - 7, mW - 7, mH - 7);
		XDrawLine(mDisplay, mWindow, mGC, 5, 5, 5, mH - 7);
		XSetLineAttributes(mDisplay, mGC, 1, LineSolid,
				CapRound, JoinRound);
	}
	return rv;
}

int Button::SetPixmap(Pixmap p, int type)
{
	ENTER;
	if (p) mDoubleBuffer = 0;
	else mDoubleBuffer = 1;
	switch (type) {
		case STATE_NORMAL:
			mNormal = p;
			break;
		case STATE_HOVER:
			mHover = p;
			break;
		case STATE_PRESSED:
			mPressed = p;
			break;
	}
	return 0;
}

int Button::SetColor(unsigned int r, unsigned int g, unsigned int b)
{
	return Widget::SetColor(r, g, b);
}

int Button::SetColor(unsigned long background)
{
	return Widget::SetColor(background);
}

int Button::SetColor(unsigned int r, unsigned int g, unsigned int b, int type)
{
	int rv = Widget::SetColor(r, g, b);
	return SetColor(mBackground, type);
}

int Button::SetColor(unsigned long background, int type)
{
	Widget::SetColor(background);
	switch (type) {
		case STATE_NORMAL:
			mNormalBackground = mBackground;
			break;
		case STATE_HOVER:
			mHoverBackground = mBackground;
			break;
		case STATE_PRESSED:
			mPressedBackground = mBackground;
			break;
	}
	return Widget::SetColor(background);
}

int Button::SetForegroundColor(unsigned int r, unsigned int g, unsigned int b)
{
	return Widget::SetForegroundColor(r, g, b);
}

int Button::SetForegroundColor(unsigned long foreground)
{
	return Widget::SetForegroundColor(foreground);
}

int Button::SetForegroundColor(unsigned int r, unsigned int g, unsigned int b,
		int type)
{
	Widget::SetForegroundColor(r, g, b);
	return SetForegroundColor(mForeground, type);
}

int Button::SetForegroundColor(unsigned long foreground, int type)
{
	Widget::SetForegroundColor(foreground);
	switch (type) {
		case STATE_NORMAL:
			mNormalForeground = mForeground;
			break;
		case STATE_HOVER:
			mHoverForeground = mForeground;
			break;
		case STATE_PRESSED:
			mPressedForeground = mForeground;
			break;
	}
	return Widget::SetForegroundColor(foreground);
}

int Button::HandleKeyRelease(XEvent *xev)
{
	char c;
	KeySym keysym;

	if (XLookupString(&(xev->xkey), &c, 1, &keysym, NULL) != NoSymbol) {
		if ((keysym==XK_Return)||(keysym==XK_KP_Enter)||(c==' ')) {
			if (mCallback) {
				TEvent tev;
				XEvent xev;
				xev.type = ButtonRelease;
				tev.type = XEVENT;
				tev.xev = &xev;
				mCallback(this, &tev, mData1, mData2);
				return 1;
			}
			return 1;
		}
	}
	return 0;
}
