#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/cursorfont.h>

#define XK_MISCELLANY
#define XK_LATIN1
#include <X11/keysymdef.h>

#include "dtk_skin.h"
#include "dtk_textentry.h"

Cursor TextEntry::sCursor = 0;

TextEntry::TextEntry(Widget *parent, int x, int y, int w, int h)
	: Widget(parent, x, y, w, h)
{
	struct timeval tv;

	mHiddenTextChar = 0;
	mBuffer = NULL;
	mBufferSize = 0;
	mBufferIndex = 0;
	mCursorIndex = 0;
	mBufferStart = 0;
	mCursorOn = 0;
	mShowCursor = 0;
	mCursorPos = 5;
	tv.tv_sec = 0;
	tv.tv_usec = 500000;
	mTextX = 2;
	mTextY = h / 2 + 4;
	mFlashingCursor = true;
	mCursorTimer = new Timer(&tv, (void*)this, NULL,
			TextEntry::CursorTimerCallback);
	mCursorTimer->Stop();
	AddChild(mCursorTimer);

	SetColor(0xffff, 0xffff, 0xffff);

	SetPixmap(Skin::GetSkinPixmap(TEXT_BACK));
	SetEventMask(ButtonPressMask | ButtonReleaseMask | ExposureMask |
			KeyPressMask | KeyReleaseMask);

	if (sCursor == 0) {
		sCursor = XCreateFontCursor(mDisplay, XC_xterm);
	}
	XDefineCursor(mDisplay, mWindow, sCursor);
	mCanHaveFocus = 1;
}

TextEntry::~TextEntry()
{
	if (mBuffer) free((void*)mBuffer);
}

int TextEntry::SetFocus()
{
	sFocusedWidget = this;
	mCursorTimer->Start();
	mShowCursor = 1;
	mCursorOn = 1;
	mDirty = 1;
	return 0;
}

int TextEntry::EventHandler(TEvent *ev)
{
	int rv = 0;

	rv = Widget::EventHandler(ev);

	if (ev == NULL) return 0;
	if (ev->type == NOEVENT) return 0;

	if (ev->type == XEVENT) {
		switch (ev->xev->type) {
			case ButtonPress:
				if (ev->xev->xbutton.window != mWindow)
					return 0;
				mCursorIndex = GetBufferPosFromCoords(
						ev->xev->xbutton.x,
						ev->xev->xbutton.y);
				SetFocus();
				break;
		}
		if ((sFocusedWidget != this) && (mShowCursor)) {
			mShowCursor = 0;
			mCursorTimer->Stop();
			mDirty = 1;
			return 0;
		}
	}
	return 0;
}

int TextEntry::Draw()
{
	int rv;

	rv = Widget::Draw();

	if (mCursorOn && mShowCursor && mEnabled)
		XDrawLine(mDisplay, mWindow, mGC, mCursorPos, 5,
				mCursorPos, mH - 4);
	return rv;
}

int TextEntry::SetBuffer(char *buffer)
{
	int i;
	if (mBuffer) free((void*)mBuffer);
	mBufferIndex = 0;
	mCursorPos = 5;
	mBufferStart = 0;
	mBuffer = NULL;
	if (buffer == NULL) {
		SetText(NULL, mTextX, mTextY);
		return 0;
	}
	for (i = 0; i < strlen(buffer); i++) {
		AddChar(buffer[i]);
	}
	if (mCallback) {
		DTKEvent dev;
		TEvent tev;
		dev.type = DTK_TEXTENTRY_CHANGED;
		dev.textentry.object = (void*)this;
		tev.type = DTKEVENT;
		tev.dev = &dev;
		mCallback(this, &tev, mData1, mData2);
	}
	return 0;
}

char *TextEntry::GetBuffer()
{
	if (mBuffer) return strdup(mBuffer);
	return 0;
}

char *TextEntry::GetBufferPtr()
{
	return mBuffer;
}

int TextEntry::GetBufferLength()
{
	if (mBuffer) return strlen(mBuffer);
	return 0;
}


int TextEntry::SetTextOffset(unsigned int offset)
{
	if (mBuffer == NULL) return 0;

	if (offset > strlen(mBuffer)) offset = strlen(mBuffer) - 1;

	mText = &mBuffer[offset];
	mBufferStart = offset;
	mDirty = 1;
	return 0;
}

int TextEntry::SetText(char *text, int x, int y)
{
	int i;
	char buf[256];

	if (mBuffer == NULL)
		return Widget::SetText(NULL, x, y);

	if (mHiddenTextChar == 0) {
		return Widget::SetText(&mBuffer[mBufferStart], x, y);
	}

	memset(buf, 0, 256);
	for (i = 0; i < 255; i++) {
		if (i > mBufferSize) return 0;
		if (i >= strlen(mBuffer)) break;
		buf[i] = mHiddenTextChar;
	}
	return Widget::SetText(buf, mTextX, mTextY);
}

int TextEntry::SetFlashingCursor(bool on)
{
	mFlashingCursor = on;
	return 0;
}

int TextEntry::DoCursor()
{
	if (mFlashingCursor) {
		mCursorOn = 1 - mCursorOn;
		mDirty = 1;
	}
	return 0;
}

int TextEntry::GetBufferPosFromCoords(int x, int y)
{
	int i, j, new_cursorpos;

	if (mBuffer == NULL) return 0;

	j = 0;
	new_cursorpos = 0;
	x -= mTextX;

	for (i = mBufferStart; i < mBufferSize; i++) {
		char s[2];
		s[0] = mBuffer[i]; s[1] = '\0';
		new_cursorpos = j;
		j += GetTextWidth(s);
		if (mBuffer[i] == '\0') break;
		if (j > x) break;
	}
	mCursorPos = new_cursorpos + mTextX + 2;
	return i;
}


int TextEntry::AddChar(char c)
{
	char s[2];

	if (mBuffer == NULL) {
		mBufferSize = 512;
		mBuffer = (char*)malloc(mBufferSize * sizeof(char));
		mBufferIndex = 0;
		mCursorIndex = 0;
	} else {
		if (mBufferIndex > mBufferSize) return 0;
		if (mCursorIndex < mBufferIndex) {
			for (int i = mBufferIndex; i > mCursorIndex; i--) {
				mBuffer[i] = mBuffer[i-1];
			}
		}
	}
	mBuffer[mCursorIndex] = c;
	mBuffer[mBufferIndex+1] = '\0';
	s[0] = c; s[1] = 0;
	mCursorPos += GetTextWidth(s);

	int i = mBufferStart;
	while (mCursorPos > mW - 4) {
		s[0] = mBuffer[i];
		mCursorPos -= GetTextWidth(s);
		i++;
	}
	mBufferStart = i;
	SetText(&mBuffer[mBufferStart], mTextX, mTextY);
	mBufferIndex++;
	mCursorIndex++;
	return 0;
}

int TextEntry::DeleteChar()
{
	if (mBuffer == NULL) return 0;
	if (mBufferIndex < 1) return 0;
	if (mCursorIndex < 1) return 0;

	if (mCursorIndex < mBufferIndex) {
		for (int i = mCursorIndex; i <= mBufferIndex; i++) {
			mBuffer[i-1] = mBuffer[i];
		}
	}
	mBufferIndex--;
	mCursorIndex--;
	mBuffer[mBufferIndex] = '\0';

	if (GetTextWidth(mBuffer) > mW - 4) {
		mBufferStart--;
		if (mBufferStart < 0) mBufferStart = 0;
	} else {
		char c = mBuffer[mCursorIndex];
		mBuffer[mCursorIndex] = '\0';
		mCursorPos = GetTextWidth(mBuffer) + 4;
		mBuffer[mCursorIndex] = c;
		mBufferStart = 0;
	}
	SetText(&mBuffer[mBufferStart], mTextX, mTextY);
	return 0;
}

int TextEntry::HandleKeyPress(XEvent *xev)
{
	char c;
	KeySym keysym;

	if (mEnabled == 0) return 0;

	if (Widget::HandleKeyPress(xev)) return 0;

	if (xev->type == ButtonRelease) return 0;
	if (xev->type == KeyRelease) return 0;

	if (XLookupString(&(xev->xkey), &c, 1, &keysym, NULL) != NoSymbol) {
		if (((unsigned char)c >= 32) && ((unsigned char)c < 128)) 
			AddChar(c);
		else if (keysym == XK_BackSpace)
			DeleteChar();
		if (mCallback) {
			DTKEvent dev;
			TEvent tev;
			dev.type = DTK_TEXTENTRY_CHANGED;
			dev.textentry.object = (void*)this;
			tev.type = DTKEVENT;
			tev.dev = &dev;
			mCallback(this, &tev, mData1, mData2);
		}
	} else if (mCallback) {
		TEvent tev;
		tev.type = XEVENT;
		tev.xev = xev;
		mCallback(this, &tev, mData1, mData2);
	}
	return 0;
}

void TextEntry::CursorTimerCallback(void *data1, void *data2)
{
	TextEntry *t = (TextEntry*)data1;
	if (t == NULL) return;
	t->DoCursor();
}

